from ._ilastik import (
    SteinbockIlastikClassificationException,
    create_and_save_ilastik_project,
    create_ilastik_crop,
    create_ilastik_image,
    fix_ilastik_project_file_inplace,
    list_ilastik_crop_files,
    list_ilastik_image_files,
    logger,
    read_ilastik_crop,
    read_ilastik_image,
    run_pixel_classification,
    try_create_ilastik_crops_from_disk,
    try_create_ilastik_images_from_disk,
    try_fix_ilastik_crops_from_disk,
    write_ilastik_crop,
    write_ilastik_image,
)

__all__ = [
    "SteinbockIlastikClassificationException",
    "create_and_save_ilastik_project",
    "create_ilastik_crop",
    "create_ilastik_image",
    "fix_ilastik_project_file_inplace",
    "list_ilastik_crop_files",
    "list_ilastik_image_files",
    "logger",
    "read_ilastik_crop",
    "read_ilastik_image",
    "run_pixel_classification",
    "try_create_ilastik_crops_from_disk",
    "try_create_ilastik_images_from_disk",
    "try_fix_ilastik_crops_from_disk",
    "write_ilastik_crop",
    "write_ilastik_image",
]



import logging
import re
from os import PathLike
from pathlib import Path
from typing import Generator, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from scipy.ndimage import maximum_filter

from .. import io
from ._preprocessing import SteinbockPreprocessingException

try:
    from readimc import MCDFile, TXTFile
    from readimc.data import Acquisition, AcquisitionBase

    imc_available = True
except:
    imc_available = False


logger = logging.getLogger(__name__)


class SteinbockIMCPreprocessingException(SteinbockPreprocessingException):
    pass


def list_mcd_files(mcd_dir: Union[str, PathLike]) -> List[Path]:
    return sorted(Path(mcd_dir).rglob("*.mcd"))


def list_txt_files(txt_dir: Union[str, PathLike]) -> List[Path]:
    return sorted(Path(txt_dir).rglob("*.txt"))


def create_panel_from_imc_panel(
    imc_panel_file: Union[str, PathLike],
    imc_panel_channel_col: str = "Metal Tag",
    imc_panel_name_col: str = "Target",
    imc_panel_keep_col: str = "full",
    imc_panel_ilastik_col: str = "ilastik",
) -> pd.DataFrame:
    imc_panel = pd.read_csv(
        imc_panel_file,
        sep=",|;",
        dtype={
            imc_panel_channel_col: pd.StringDtype(),
            imc_panel_name_col: pd.StringDtype(),
            imc_panel_keep_col: pd.BooleanDtype(),
            imc_panel_ilastik_col: pd.BooleanDtype(),
        },
        engine="python",
        true_values=["1"],
        false_values=["0"],
    )
    for required_col in (imc_panel_channel_col, imc_panel_name_col):
        if required_col not in imc_panel:
            raise SteinbockIMCPreprocessingException(
                f"Missing '{required_col}' column in IMC panel"
            )
    for notnan_col in (
        imc_panel_channel_col,
        imc_panel_keep_col,
        imc_panel_ilastik_col,
    ):
        if notnan_col in imc_panel and imc_panel[notnan_col].isna().any():
            raise SteinbockIMCPreprocessingException(
                f"Missing values for '{notnan_col}' in IMC panel"
            )
    rename_columns = {
        imc_panel_channel_col: "channel",
        imc_panel_name_col: "name",
        imc_panel_keep_col: "keep",
        imc_panel_ilastik_col: "ilastik",
    }
    drop_columns = [
        panel_col
        for imc_panel_col, panel_col in rename_columns.items()
        if panel_col in imc_panel.columns and panel_col != imc_panel_col
    ]
    panel = imc_panel.drop(columns=drop_columns).rename(columns=rename_columns)
    for _, g in panel.groupby("channel"):
        panel.loc[g.index, "name"] = " / ".join(g["name"].dropna().unique())
        if "keep" in panel:
            panel.loc[g.index, "keep"] = g["keep"].any()
        if "ilastik" in panel:
            panel.loc[g.index, "ilastik"] = g["ilastik"].any()
    panel = panel.groupby(panel["channel"].values).aggregate("first")
    panel = _clean_panel(panel)  # ilastik column may be nullable uint8 now
    ilastik_mask = panel["ilastik"].fillna(False).astype(bool)
    panel["ilastik"] = pd.Series(dtype=pd.UInt8Dtype())
    panel.loc[ilastik_mask, "ilastik"] = range(1, ilastik_mask.sum() + 1)
    return panel


def create_panel_from_mcd_files(
    mcd_files: Sequence[Union[str, PathLike]]
) -> pd.DataFrame:
    panels = []
    for mcd_file in mcd_files:
        with MCDFile(mcd_file) as f:
            for slide in f.slides:
                for acquisition in slide.acquisitions:
                    panel = pd.DataFrame(
                        data={
                            "channel": pd.Series(
                                data=acquisition.channel_names,
                                dtype=pd.StringDtype(),
                            ),
                            "name": pd.Series(
                                data=acquisition.channel_labels,
                                dtype=pd.StringDtype(),
                            ),
                        },
                    )
                    panels.append(panel)
    panel = pd.concat(panels, ignore_index=True, copy=False)
    panel.drop_duplicates(inplace=True, ignore_index=True)
    return _clean_panel(panel)


def create_panel_from_txt_files(
    txt_files: Sequence[Union[str, PathLike]]
) -> pd.DataFrame:
    panels = []
    for txt_file in txt_files:
        with TXTFile(txt_file) as f:
            panel = pd.DataFrame(
                data={
                    "channel": pd.Series(data=f.channel_names, dtype=pd.StringDtype()),
                    "name": pd.Series(data=f.channel_labels, dtype=pd.StringDtype()),
                },
            )
            panels.append(panel)
    panel = pd.concat(panels, ignore_index=True, copy=False)
    panel.drop_duplicates(inplace=True, ignore_index=True)
    return _clean_panel(panel)


def filter_hot_pixels(img: np.ndarray, thres: float) -> np.ndarray:
    kernel = np.ones((1, 3, 3), dtype=bool)
    kernel[0, 1, 1] = False
    max_neighbor_img = maximum_filter(img, footprint=kernel, mode="mirror")
    return np.where(img - max_neighbor_img > thres, max_neighbor_img, img)


def preprocess_image(img: np.ndarray, hpf: Optional[float] = None) -> np.ndarray:
    img = img.astype(np.float32)
    if hpf is not None:
        img = filter_hot_pixels(img, hpf)
    return io._to_dtype(img, io.img_dtype)


def try_preprocess_images_from_disk(
    mcd_files: Sequence[Union[str, PathLike]],
    txt_files: Sequence[Union[str, PathLike]],
    channel_names: Optional[Sequence[str]] = None,
    hpf: Optional[float] = None,
) -> Generator[
    Tuple[Path, Optional["Acquisition"], np.ndarray, Optional[Path], bool],
    None,
    None,
]:
    unmatched_txt_files = list(txt_files)
    # process mcd files in descending order to avoid ambiguous txt file matching
    # see https://github.com/BodenmillerGroup/steinbock/issues/100
    for mcd_file in sorted(
        mcd_files, key=lambda mcd_file: Path(mcd_file).stem, reverse=True
    ):
        try:
            with MCDFile(mcd_file) as f_mcd:
                for slide in f_mcd.slides:
                    for acquisition in slide.acquisitions:
                        matched_txt_file = _match_txt_file(
                            mcd_file, acquisition, unmatched_txt_files
                        )
                        if matched_txt_file is not None:
                            unmatched_txt_files.remove(matched_txt_file)
                        channel_ind = None
                        if channel_names is not None:
                            channel_ind = _get_channel_indices(
                                acquisition, channel_names
                            )
                            if isinstance(channel_ind, str):
                                logger.warning(
                                    f"Channel {channel_ind} not found for "
                                    f"acquisition {acquisition.id} in file "
                                    "{mcd_file}; skipping acquisition"
                                )
                                continue
                        img = None
                        recovered = False
                        try:
                            img = f_mcd.read_acquisition(acquisition)
                        except IOError as e:
                            logger.warning(
                                f"Error reading acquisition {acquisition.id} "
                                f"from file {mcd_file}: {e}"
                            )
                            if matched_txt_file is not None:
                                logger.warning(
                                    f"Restoring from file {matched_txt_file}"
                                )
                                try:
                                    with TXTFile(matched_txt_file) as f_txt:
                                        img = f_txt.read_acquisition()
                                        if channel_names is not None:
                                            channel_ind = _get_channel_indices(
                                                f_txt, channel_names
                                            )
                                            if isinstance(channel_ind, str):
                                                logger.warning(
                                                    f"Channel {channel_ind} "
                                                    "not found in file "
                                                    f"{matched_txt_file}; "
                                                    "skipping acquisition"
                                                )
                                                continue
                                    recovered = True
                                except IOError as e2:
                                    logger.error(
                                        f"Error reading file {matched_txt_file}: {e2}"
                                    )
                        if img is not None:  # exceptions ...
                            if channel_ind is not None:
                                img = img[channel_ind, :, :]
                            img = preprocess_image(img, hpf=hpf)
                            yield (
                                Path(mcd_file),
                                acquisition,
                                img,
                                Path(matched_txt_file)
                                if matched_txt_file is not None
                                else None,
                                recovered,
                            )
                            del img
        except:
            logger.exception(f"Error reading file {mcd_file}")
    while len(unmatched_txt_files) > 0:
        txt_file = unmatched_txt_files.pop(0)
        try:
            channel_ind = None
            with TXTFile(txt_file) as f:
                if channel_names is not None:
                    channel_ind = _get_channel_indices(f, channel_names)
                    if isinstance(channel_ind, str):
                        logger.warning(
                            f"Channel {channel_ind} not found in file "
                            f"{txt_file}; skipping acquisition"
                        )
                        continue
                img = f.read_acquisition()
            if channel_ind is not None:
                img = img[channel_ind, :, :]
            img = preprocess_image(img, hpf=hpf)
            yield Path(txt_file), None, img, None, False
            del img
        except:
            logger.exception(f"Error reading file {txt_file}")


def _clean_panel(panel: pd.DataFrame) -> pd.DataFrame:
    panel.sort_values(
        "channel",
        key=lambda s: pd.to_numeric(s.str.replace("[^0-9]", "", regex=True)),
        inplace=True,
    )
    name_dupl_mask = panel["name"].duplicated(keep=False)
    name_suffixes = panel.groupby("name").cumcount().map(lambda i: f" {i + 1}")
    panel.loc[name_dupl_mask, "name"] += name_suffixes[name_dupl_mask]
    if "keep" not in panel:
        panel["keep"] = pd.Series(dtype=pd.BooleanDtype())
        panel.loc[:, "keep"] = True
    if "ilastik" not in panel:
        panel["ilastik"] = pd.Series(dtype=pd.UInt8Dtype())
        panel.loc[panel["keep"], "ilastik"] = range(1, panel["keep"].sum() + 1)
    if "deepcell" not in panel:
        panel["deepcell"] = pd.Series(dtype=pd.UInt8Dtype())
    next_column_index = 0
    for column in ("channel", "name", "keep", "ilastik", "deepcell"):
        if column in panel:
            column_data = panel[column]
            panel.drop(columns=[column], inplace=True)
            panel.insert(next_column_index, column, column_data)
            next_column_index += 1
    return panel


def _match_txt_file(
    mcd_file: Union[str, PathLike],
    acquisition: Acquisition,
    txt_files: Sequence[Union[str, PathLike]],
) -> Union[str, PathLike, None]:
    txt_file_name_pattern = re.compile(
        rf"{Path(mcd_file).stem}.*_0*{acquisition.id}.txt"
    )
    filtered_txt_files = [
        txt_file
        for txt_file in txt_files
        if txt_file_name_pattern.match(Path(txt_file).name)
    ]
    if len(filtered_txt_files) == 1:
        return filtered_txt_files[0]
    if len(filtered_txt_files) > 1:
        logger.warning(
            "Ambiguous txt file matching for %s: %s; continuing without a match",
            mcd_file,
            ", ".join(filtered_txt_files),
        )
    return None


def _get_channel_indices(
    acquisition: AcquisitionBase, channel_names: Sequence[str]
) -> Union[List[int], str]:
    channel_indices = []
    for channel_name in channel_names:
        if channel_name not in acquisition.channel_names:
            return channel_name
        channel_indices.append(acquisition.channel_names.index(channel_name))
    return channel_indices



from pathlib import Path

import click
import click_log
import networkx as nx

from ... import io
from ..._cli.utils import catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import graphs


@click.command(name="graphs", help="Export neighbors as spatial object graphs")
@click.option(
    "--neighbors",
    "neighbors_dir",
    type=click.Path(exists=True, file_okay=False),
    default="neighbors",
    show_default=True,
    help="Path to the neighbors directory",
)
@click.option(
    "--data",
    "data_dirs",
    multiple=True,
    type=click.STRING,
    help="Object data (e.g. intensities, regionprops) to use as attributes",
)
@click.option(
    "--format",
    "graph_format",
    type=click.Choice(["graphml", "gexf", "gml"], case_sensitive=False),
    default="graphml",
    show_default=True,
    help="AnnData file format to use",
)
@click.option(
    "-o",
    "graph_dir",
    type=click.Path(file_okay=False),
    default="graphs",
    show_default=True,
    help="Path to the networkx output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def graphs_cmd(neighbors_dir, data_dirs, graph_format, graph_dir):
    neighbors_files = io.list_neighbors_files(neighbors_dir)
    data_file_lists = [
        io.list_data_files(data_dir, base_files=neighbors_files)
        for data_dir in data_dirs
    ]
    Path(graph_dir).mkdir(exist_ok=True)
    for (
        neighbors_file,
        data_files,
        graph,
    ) in graphs.try_convert_to_networkx_from_disk(neighbors_files, *data_file_lists):
        graph_file = Path(graph_dir) / neighbors_file.name
        if graph_format == "graphml":
            graph_file = io._as_path_with_suffix(graph_file, ".graphml")
            nx.write_graphml(graph, str(graph_file))
        elif graph_format == "gexf":
            graph_file = io._as_path_with_suffix(graph_file, ".gexf")
            nx.write_gexf(graph, str(graph_file))
        elif graph_format == "gml":
            graph_file = io._as_path_with_suffix(graph_file, ".gml")
            nx.write_gml(graph, str(graph_file))
        else:
            raise NotImplementedError()
        logger.info(graph_file)



from pathlib import Path

import click
import click_log
import numpy as np
import pandas as pd

from ... import io
from ..._cli.utils import OrderedClickGroup, catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import external


@click.group(
    name="external",
    cls=OrderedClickGroup,
    help="Preprocess external image data",
)
def external_cmd_group():
    pass


@external_cmd_group.command(
    name="panel", help="Create a panel from external image data"
)
@click.option(
    "--img",
    "ext_img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="external",
    show_default=True,
    help="Path to the external image file directory",
)
@click.option(
    "-o",
    "panel_file",
    type=click.Path(dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel output file",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def panel_cmd(ext_img_dir, panel_file):
    ext_img_files = external.list_image_files(ext_img_dir)
    panel = external.create_panel_from_image_files(ext_img_files)
    io.write_panel(panel, panel_file)
    logger.info(panel_file)


@external_cmd_group.command(name="images", help="Extract external images")
@click.option(
    "--img",
    "ext_img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="external",
    show_default=True,
    help="Path to the external image file directory",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the steinbock panel file",
)
@click.option(
    "--mmap/--no-mmap",
    "mmap",
    default=False,
    show_default=True,
    help="Use memory mapping for writing images",
)
@click.option(
    "--imgout",
    "img_dir",
    type=click.Path(file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image output directory",
)
@click.option(
    "--infoout",
    "image_info_file",
    type=click.Path(dir_okay=False),
    default="images.csv",
    show_default=True,
    help="Path to the image information output file",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def images_cmd(ext_img_dir, panel_file, mmap, img_dir, image_info_file):
    channel_indices = None
    if Path(panel_file).exists():
        panel = io.read_panel(panel_file)
        if "channel" in panel:
            channel_indices = panel["channel"].astype(int).sub(1).tolist()
    ext_img_files = external.list_image_files(ext_img_dir)
    image_info_data = []
    Path(img_dir).mkdir(exist_ok=True)
    for ext_img_file, img in external.try_preprocess_images_from_disk(ext_img_files):
        # filter channels here rather than in try_preprocess_images_from_disk,
        # to avoid advanced indexing creating a copy of img (relevant for mmap)
        if channel_indices is not None:
            if max(channel_indices) > img.shape[0]:
                logger.warning(
                    f"Channel indices out of bounds for file {ext_img_file} "
                    f"with {img.shape[0]} channels"
                )
                continue
            cur_channel_indices = channel_indices
        else:
            cur_channel_indices = list(range(img.shape[0]))
        img_file = io._as_path_with_suffix(
            Path(img_dir) / Path(ext_img_file).name, ".tiff"
        )
        out_shape = list(img.shape)
        out_shape[0] = len(cur_channel_indices)
        out_shape = tuple(out_shape)
        if mmap:
            out = io.mmap_image(img_file, mode="r+", shape=out_shape, dtype=img.dtype)
        else:
            out = np.empty(out_shape, dtype=img.dtype)
        for i, channel_index in enumerate(cur_channel_indices):
            out[i, :, :] = img[channel_index, :, :]
            if mmap:
                out.flush()
        if not mmap:
            io.write_image(out, img_file)
        image_info_row = {
            "image": img_file.name,
            "width_px": img.shape[2],
            "height_px": img.shape[1],
            "num_channels": img.shape[0],
        }
        image_info_data.append(image_info_row)
        logger.info(img_file)
        del img
    image_info = pd.DataFrame(data=image_info_data)
    io.write_image_info(image_info, image_info_file)



import click

from .matching import match_cmd
from .mosaics import mosaics_cmd_group
from ..._cli.utils import OrderedClickGroup


@click.group(name="utils", cls=OrderedClickGroup, help="Various utilities and tools")
def utils_cmd_group():
    pass


utils_cmd_group.add_command(match_cmd)
utils_cmd_group.add_command(mosaics_cmd_group)



from pathlib import Path

import numpy as np
from steinbock import io
from steinbock.utils import matching


class TestMatchingUtils:
    def test_match_masks(self):
        mask1 = np.array(
            [
                [1, 1, 0],
                [0, 0, 0],
                [0, 2, 2],
            ],
            dtype=io.mask_dtype,
        )
        mask2 = np.array(
            [
                [3, 3, 0],
                [0, 0, 0],
                [0, 4, 4],
            ],
            dtype=io.mask_dtype,
        )
        matching.match_masks(mask1, mask2)  # TODO

    def test_try_match_masks_from_disk(self, imc_test_data_steinbock_path: Path):
        mask_files1 = io.list_mask_files(imc_test_data_steinbock_path / "masks")
        mask_files2 = io.list_mask_files(
            imc_test_data_steinbock_path / "masks", base_files=mask_files1
        )
        gen = matching.try_match_masks_from_disk(mask_files1, mask_files2)
        for mask_file1, mask_file2, df in gen:
            pass  # TODO



import shutil
import subprocess
from importlib import resources
from os import PathLike
from pathlib import Path
from typing import Union

from ..._env import run_captured
from . import data as cellprofiler_data


def create_and_save_segmentation_pipeline(
    segmentation_pipeline_file: Union[str, PathLike]
) -> None:
    with resources.open_binary(cellprofiler_data, "cell_segmentation.cppipe") as fsrc:
        with open(segmentation_pipeline_file, mode="wb") as fdst:
            shutil.copyfileobj(fsrc, fdst)


def try_segment_objects(
    cellprofiler_binary: Union[str, PathLike],
    segmentation_pipeline_file: Union[str, PathLike],
    probabilities_dir: Union[str, PathLike],
    mask_dir: Union[str, PathLike],
    cellprofiler_plugin_dir: Union[str, PathLike, None] = None,
) -> subprocess.CompletedProcess:
    args = [
        str(cellprofiler_binary),
        "-c",
        "-r",
        "-p",
        str(segmentation_pipeline_file),
        "-i",
        str(probabilities_dir),
        "-o",
        str(mask_dir),
    ]
    if cellprofiler_plugin_dir is not None and Path(cellprofiler_plugin_dir).exists():
        args.append(f"--plugins-directory={cellprofiler_plugin_dir}")
    return run_captured(args)



import logging

import click_log

from ._cli import steinbock_cmd_group
from ._steinbock import logger

# click_log.basic_config(logger=logger)
logger_handler = click_log.ClickHandler()
logger_handler.formatter = logging.Formatter(
    fmt="%(asctime)s %(levelname)s %(name)s - %(message)s"
)
logger.handlers = [logger_handler]
logger.propagate = False

if __name__ == "__main__":
    steinbock_cmd_group(prog_name="steinbock")



import json
import logging
import shutil
import subprocess
from enum import IntEnum
from importlib import resources
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Callable,
    Generator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from uuid import uuid1

import cv2
import h5py
import numpy as np

from ... import io
from ..._env import run_captured
from .._classification import SteinbockClassificationException
from . import data as ilastik_data

logger = logging.getLogger(__name__.rpartition(".")[0])


class SteinbockIlastikClassificationException(SteinbockClassificationException):
    pass


class _VigraAxisInfo(IntEnum):
    CHANNELS = 1
    SPACE = 2
    ANGLE = 4
    TIME = 8
    FREQUENCY = 16
    EDGE = 32
    UNKNOWN_AXIS_TYPE = 64


_img_dataset_path = "img"
_crop_dataset_path = "crop"
_dataset_display_mode = "grayscale"
_dataset_axistags = json.dumps(
    {
        "axes": [
            {"key": "c", "typeFlags": _VigraAxisInfo.CHANNELS.value},
            {"key": "y", "typeFlags": _VigraAxisInfo.SPACE.value},
            {"key": "x", "typeFlags": _VigraAxisInfo.SPACE.value},
        ]
    }
)
_h5py_libver = "earliest"


def list_ilastik_image_files(ilastik_img_dir: Union[str, PathLike]) -> List[Path]:
    return sorted(Path(ilastik_img_dir).rglob("*.h5"))


def list_ilastik_crop_files(ilastik_crop_dir: Union[str, PathLike]) -> List[Path]:
    return sorted(Path(ilastik_crop_dir).rglob("*.h5"))


def read_ilastik_image(ilastik_img_file: Union[str, PathLike]) -> np.ndarray:
    with h5py.File(ilastik_img_file, mode="r", libver=_h5py_libver) as f:
        return io._to_dtype(f[str(_img_dataset_path)][()], io.img_dtype)


def read_ilastik_crop(ilastik_crop_file: Union[str, PathLike]) -> np.ndarray:
    with h5py.File(ilastik_crop_file, mode="r", libver=_h5py_libver) as f:
        return io._to_dtype(f[str(_crop_dataset_path)][()], io.img_dtype)


def write_ilastik_image(
    ilastik_img: np.ndarray, ilastik_img_file: Union[str, PathLike]
) -> None:
    ilastik_img = io._to_dtype(ilastik_img, io.img_dtype)
    with h5py.File(ilastik_img_file, mode="w", libver=_h5py_libver) as f:
        dataset = _create_or_replace_dataset(f, _img_dataset_path, ilastik_img)
        dataset.attrs["display_mode"] = _str_encode(_dataset_display_mode, ascii=True)
        dataset.attrs["axistags"] = _str_encode(_dataset_axistags, ascii=True)
        dataset.attrs["steinbock"] = True


def write_ilastik_crop(
    ilastik_crop: np.ndarray, ilastik_crop_file: Union[str, PathLike]
) -> None:
    ilastik_crop = io._to_dtype(ilastik_crop, io.img_dtype)
    with h5py.File(ilastik_crop_file, mode="w", libver=_h5py_libver) as f:
        dataset = _create_or_replace_dataset(f, _crop_dataset_path, ilastik_crop)
        dataset.attrs["display_mode"] = _str_encode(_dataset_display_mode, ascii=True)
        dataset.attrs["axistags"] = _str_encode(_dataset_axistags, ascii=True)
        dataset.attrs["steinbock"] = True


def create_ilastik_image(
    img: np.ndarray,
    channel_groups: Optional[np.ndarray] = None,
    aggr_func: Callable[[np.ndarray], np.ndarray] = np.mean,
    prepend_mean: bool = True,
    mean_factor: float = 100.0,
    scale_factor: int = 1,
) -> np.ndarray:
    ilastik_img = img
    if channel_groups is not None:
        ilastik_img = np.stack(
            [
                aggr_func(ilastik_img[channel_groups == channel_group], axis=0)
                for channel_group in np.unique(channel_groups)
                if not np.isnan(channel_group)
            ]
        )
    if prepend_mean:
        mean_img = ilastik_img.mean(axis=0, keepdims=True) * mean_factor
        ilastik_img = np.concatenate((mean_img, ilastik_img))
    if scale_factor > 1:
        # bilinear resizing (for compatibility with IMC Segmentation Pipeline)
        # # use OpenCV instead of scikit-image for memory reasons
        # output_shape = (
        #     ilastik_img.shape[0],
        #     ilastik_img.shape[1] * scale_factor,
        #     ilastik_img.shape[2] * scale_factor,
        # )
        # ilastik_img = resize(
        #     ilastik_img, output_shape, order=1, mode="symmetric"
        # )
        ilastik_img = np.stack(
            [
                cv2.resize(
                    ilastik_channel_img,
                    None,
                    fx=scale_factor,
                    fy=scale_factor,
                    interpolation=cv2.INTER_LINEAR,
                )
                for ilastik_channel_img in ilastik_img
            ]
        )

    return io._to_dtype(ilastik_img, io.img_dtype)


def try_create_ilastik_images_from_disk(
    img_files: Sequence[Union[str, PathLike]],
    channel_groups: Optional[np.ndarray] = None,
    aggr_func: Callable[[np.ndarray], np.ndarray] = np.mean,
    prepend_mean: bool = True,
    mean_factor: float = 100.0,
    scale_factor: int = 1,
) -> Generator[Tuple[Path, np.ndarray], None, None]:
    for img_file in img_files:
        try:
            ilastik_img = create_ilastik_image(
                io.read_image(img_file, native_dtype=True),
                channel_groups=channel_groups,
                aggr_func=aggr_func,
                prepend_mean=prepend_mean,
                mean_factor=mean_factor,
                scale_factor=scale_factor,
            )
            yield Path(img_file), ilastik_img
            del ilastik_img
        except:
            logger.exception(f"Error creating Ilastik image from file {img_file}")


def create_ilastik_crop(
    ilastik_img: np.ndarray, ilastik_crop_size: int, rng: np.random.Generator
) -> Tuple[Optional[int], Optional[int], Optional[np.ndarray]]:
    if all(shape >= ilastik_crop_size for shape in ilastik_img.shape[1:]):
        ilastik_crop_x = 0
        if ilastik_img.shape[2] > ilastik_crop_size:
            ilastik_crop_x = rng.integers(ilastik_img.shape[2] - ilastik_crop_size)
        ilastik_crop_y = 0
        if ilastik_img.shape[1] > ilastik_crop_size:
            ilastik_crop_y = rng.integers(ilastik_img.shape[1] - ilastik_crop_size)
        ilastik_crop = ilastik_img[
            :,
            ilastik_crop_y : (ilastik_crop_y + ilastik_crop_size),
            ilastik_crop_x : (ilastik_crop_x + ilastik_crop_size),
        ]
        return (
            ilastik_crop_x,
            ilastik_crop_y,
            io._to_dtype(ilastik_crop, io.img_dtype),
        )
    return None, None, None


def try_create_ilastik_crops_from_disk(
    ilastik_img_files: Sequence[Union[str, PathLike]],
    ilastik_crop_size: int,
    rng: np.random.Generator,
) -> Generator[
    Tuple[Path, Optional[int], Optional[int], Optional[np.ndarray]],
    None,
    None,
]:
    for ilastik_img_file in ilastik_img_files:
        try:
            ilastik_crop_x, ilastik_crop_y, ilastik_crop = create_ilastik_crop(
                read_ilastik_image(ilastik_img_file), ilastik_crop_size, rng
            )
            yield Path(ilastik_img_file), ilastik_crop_x, ilastik_crop_y, ilastik_crop
            del ilastik_crop
        except:
            logger.exception(
                f"Error creating Ilastik crop from file {ilastik_img_file}"
            )


def create_and_save_ilastik_project(
    ilastik_crop_files: Sequence[Union[str, PathLike]],
    ilastik_project_file: Union[str, PathLike],
) -> None:
    dataset_id = str(uuid1())
    with resources.open_binary(ilastik_data, "pixel_classifier.ilp") as fsrc:
        with open(ilastik_project_file, mode="wb") as fdst:
            shutil.copyfileobj(fsrc, fdst)
    with h5py.File(ilastik_project_file, mode="a", libver=_h5py_libver) as f:
        infos_group = f["Input Data/infos"]
        for i, ilastik_crop_file in enumerate(ilastik_crop_files):
            rel_ilastik_crop_file = Path(ilastik_crop_file).relative_to(
                Path(ilastik_project_file).parent
            )
            with h5py.File(ilastik_crop_file, mode="r", libver=_h5py_libver) as f_crop:
                ilastik_crop_shape = f_crop[_crop_dataset_path].shape
            lane_group = infos_group.create_group(f"lane{i:04d}")
            lane_group.create_group("Prediction Mask")
            raw_data_group = lane_group.create_group("Raw Data")
            _fix_raw_data_group_inplace(
                raw_data_group,
                dataset_id=dataset_id,
                file_path=str(rel_ilastik_crop_file / _crop_dataset_path),
                nickname=Path(ilastik_crop_file).stem,
                shape=ilastik_crop_shape,
            )


def run_pixel_classification(
    ilastik_binary: Union[str, PathLike],
    ilastik_project_file: Union[str, PathLike],
    ilastik_img_files: Sequence[Union[str, PathLike]],
    ilastik_probab_dir: Union[str, PathLike],
    num_threads: Optional[int] = None,
    memory_limit: Optional[int] = None,
    ilastik_env: Optional[Mapping[str, str]] = None,
) -> subprocess.CompletedProcess:
    output_filename_format = Path(ilastik_probab_dir) / "{nickname}.tiff"
    args = [
        str(ilastik_binary),
        "--headless",
        f"--project={ilastik_project_file}",
        "--readonly",
        "--input_axes=cyx",
        "--export_source=Probabilities",
        "--output_format=tiff",
        f"--output_filename_format={output_filename_format}",
        "--export_dtype=uint16",
        "--output_axis_order=yxc",
        "--pipeline_result_drange=(0.0,1.0)",
        "--export_drange=(0,65535)",
    ]
    for ilastik_img_file in ilastik_img_files:
        args.append(str(Path(ilastik_img_file) / _img_dataset_path))
    if ilastik_env is not None:
        ilastik_env = ilastik_env.copy()
        if num_threads is not None:
            ilastik_env["LAZYFLOW_THREADS"] = num_threads
        if memory_limit is not None:
            ilastik_env["LAZYFLOW_TOTAL_RAM_MB"] = memory_limit
    result = run_captured(args, env=ilastik_env)
    ilastik_probab_files = Path(ilastik_probab_dir).rglob(f"*-{_img_dataset_path}.tiff")
    for ilastik_probab_file in sorted(ilastik_probab_files):
        ilastik_probab_file.rename(
            ilastik_probab_file.with_name(
                ilastik_probab_file.name.replace(f"-{_img_dataset_path}", "")
            )
        )
    return result


def try_fix_ilastik_crops_from_disk(
    ilastik_crop_files: Sequence[Union[str, PathLike]],
    orig_axis_order: Union[str, Sequence, None] = None,
) -> Generator[Tuple[Path, Tuple[int, ...], np.ndarray], None, None]:
    for ilastik_crop_file in ilastik_crop_files:
        try:
            with h5py.File(ilastik_crop_file, mode="r", libver=_h5py_libver) as f:
                ilastik_crop_dataset = None
                if _crop_dataset_path in f:
                    ilastik_crop_dataset = f[_crop_dataset_path]
                elif Path(ilastik_crop_file).stem in f:
                    ilastik_crop_dataset = f[Path(ilastik_crop_file).stem]
                elif len(f) == 1:
                    ilastik_crop_dataset = next(iter(f.values()))
                else:
                    raise SteinbockIlastikClassificationException(
                        f"Unknown dataset: {ilastik_crop_file}"
                    )
                if ilastik_crop_dataset.attrs.get("steinbock", False):
                    continue
                ilastik_crop = ilastik_crop_dataset[()]
                if orig_axis_order is not None:
                    orig_axis_order = list(orig_axis_order)
                elif "axistags" in ilastik_crop_dataset.attrs:
                    axis_tags, _ = _str_decode(ilastik_crop_dataset.attrs["axistags"])
                    axis_tags_json = json.loads(axis_tags)
                    orig_axis_order = [
                        a_json["key"] for a_json in axis_tags_json["axes"]
                    ]
                else:
                    raise SteinbockIlastikClassificationException(
                        f"Unknown axis order: {ilastik_crop_file}"
                    )
            if len(orig_axis_order) != ilastik_crop.ndim:
                raise SteinbockIlastikClassificationException(
                    f"Incompatible axis order: {ilastik_crop_file}"
                )
            channel_axis_index = orig_axis_order.index("c")
            num_channels = ilastik_crop.size // (
                ilastik_crop.shape[orig_axis_order.index("x")]
                * ilastik_crop.shape[orig_axis_order.index("y")]
            )
            if ilastik_crop.shape[channel_axis_index] != num_channels:
                channel_axis_indices = (
                    i
                    for i, a in enumerate(orig_axis_order)
                    if ilastik_crop.shape[i] == num_channels and a not in ("x", "y")
                )
                channel_axis_index = next(channel_axis_indices, None)
            if channel_axis_index is None:
                raise SteinbockIlastikClassificationException(
                    f"Unknown channel axis: {ilastik_crop_file}"
                )
            axis_order = orig_axis_order.copy()
            axis_order.insert(0, axis_order.pop(channel_axis_index))
            axis_order.insert(1, axis_order.pop(axis_order.index("y")))
            axis_order.insert(2, axis_order.pop(axis_order.index("x")))
            transpose_axes = [orig_axis_order.index(a) for a in axis_order]
            ilastik_crop = np.transpose(ilastik_crop, axes=transpose_axes)
            ilastik_crop = np.reshape(ilastik_crop, ilastik_crop.shape[:3])
            ilastik_crop = io._to_dtype(ilastik_crop, io.img_dtype)
            yield Path(ilastik_crop_file), transpose_axes, ilastik_crop
            del ilastik_crop
        except:
            logger.exception(f"Error fixing Ilastik crop from file {ilastik_crop_file}")


def fix_ilastik_project_file_inplace(
    ilastik_project_file: Union[str, PathLike],
    ilastik_crop_dir: Union[str, PathLike],
    ilastik_probab_dir: Union[str, PathLike],
    ilastik_crop_shapes: Mapping[str, Tuple[int, ...]],
    transpose_axes: Sequence[int],
) -> None:
    rel_ilastik_crop_dir = Path(ilastik_crop_dir).relative_to(
        Path(ilastik_project_file).parent
    )
    rel_ilastik_probab_dir = Path(ilastik_probab_dir).relative_to(
        Path(ilastik_project_file).parent
    )
    with h5py.File(ilastik_project_file, "a", libver=_h5py_libver) as f:
        input_data_group = f.get("Input Data")
        if input_data_group is not None:
            _fix_input_data_group_inplace(
                input_data_group, rel_ilastik_crop_dir, ilastik_crop_shapes
            )
        pixel_classification_group = f.get("PixelClassification")
        if pixel_classification_group is not None:
            _fix_pixel_classification_group_inplace(
                pixel_classification_group, transpose_axes
            )
        prediction_export_group = f["Prediction Export"]
        if prediction_export_group is not None:
            _fix_prediction_export_group_inplace(
                prediction_export_group, rel_ilastik_probab_dir
            )


def _fix_input_data_group_inplace(
    input_data_group: h5py.Group,
    rel_ilastik_crop_dir: Path,
    ilastik_crop_shapes: Mapping[str, Tuple[int, ...]],
) -> None:
    infos_group = input_data_group.get("infos")
    if infos_group is not None:
        for lane_group in infos_group.values():
            raw_data_group = lane_group.get("Raw Data")
            if raw_data_group is not None:
                file_path, _ = _str_decode(raw_data_group["filePath"][()])
                ilastik_crop_file = _get_hdf5_file(file_path)
                _fix_raw_data_group_inplace(
                    raw_data_group,
                    file_path=str(
                        rel_ilastik_crop_dir
                        / ilastik_crop_file.name
                        / _crop_dataset_path
                    ),
                    nickname=ilastik_crop_file.stem,
                    shape=ilastik_crop_shapes[ilastik_crop_file.stem],
                )
    local_data_group = input_data_group.get("local_data")
    if local_data_group is not None:
        local_data_group.clear()


def _fix_pixel_classification_group_inplace(
    pixel_classification_group: h5py.Group, transpose_axes: List[int]
) -> None:
    label_sets_group = pixel_classification_group.get("LabelSets")
    if label_sets_group is not None:
        for labels_group in label_sets_group.values():
            block_dataset_names = list(labels_group.keys())
            for block_dataset_name in block_dataset_names:
                block_dataset = labels_group[block_dataset_name]
                block = block_dataset[()]
                block = np.transpose(block, axes=transpose_axes)
                block = np.reshape(block, block.shape[:3])
                block_slice, block_slice_ascii = _str_decode(
                    block_dataset.attrs["blockSlice"]
                )
                block_slice = block_slice[1:-1].split(",")
                block_slice = [block_slice[i] for i in transpose_axes[:3]]
                block_slice = f"[{','.join(block_slice)}]"
                del labels_group[block_dataset_name]
                block_dataset = _create_or_replace_dataset(
                    labels_group, block_dataset_name, block
                )
                block_dataset.attrs["blockSlice"] = _str_encode(
                    block_slice, ascii=block_slice_ascii
                )


def _fix_prediction_export_group_inplace(
    prediction_export_group: h5py.Group, rel_ilastik_probab_dir: Path
) -> None:
    logger.warning(
        "When exporting data from the graphical user interface of Ilastik, "
        "please manually edit the export image settings and enable "
        "renormalizing [min, max] from [0.0, 1.0] to [0, 65535]"
    )
    # _create_or_replace_dataset(
    #   prediction_export_group,
    #   "InputMin",
    #   np.array([0.0], dtype=np.float32),
    # )
    # _create_or_replace_dataset(
    #   prediction_export_group,
    #   "InputMax", np.array([1.0], dtype=np.float32),
    # )
    _create_or_replace_dataset(
        prediction_export_group, "ExportMin", np.array([0], dtype=np.uint16)
    )
    _create_or_replace_dataset(
        prediction_export_group,
        "ExportMax",
        np.array([65535], dtype=np.uint16),
    )
    _create_or_replace_dataset(
        prediction_export_group, "ExportDtype", "uint16", ascii=False
    )
    _create_or_replace_dataset(
        prediction_export_group, "OutputAxisOrder", "yxc", ascii=True
    )
    _create_or_replace_dataset(
        prediction_export_group, "OutputFormat", "tiff", ascii=True
    )
    _create_or_replace_dataset(
        prediction_export_group,
        "OutputFilenameFormat",
        f"{rel_ilastik_probab_dir}/{{nickname}}.tiff",
        ascii=True,
    )
    _create_or_replace_dataset(
        prediction_export_group,
        "OutputInternalPath",
        "exported_data",
        ascii=True,
    )
    _create_or_replace_dataset(
        prediction_export_group, "StorageVersion", "0.1", ascii=False
    )


def _fix_raw_data_group_inplace(
    raw_data_group: h5py.Group,
    dataset_id: Optional[str] = None,
    file_path: Optional[str] = None,
    nickname: Optional[str] = None,
    shape: Optional[Tuple[int, ...]] = None,
) -> None:
    _create_or_replace_dataset(
        raw_data_group,
        "__class__",
        "RelativeFilesystemDatasetInfo",
        ascii=True,
    )
    _create_or_replace_dataset(raw_data_group, "allowLabels", True)
    _create_or_replace_dataset(
        raw_data_group, "axistags", _dataset_axistags, ascii=True
    )
    if dataset_id is not None:
        _create_or_replace_dataset(raw_data_group, "datasetId", dataset_id, ascii=True)
    _create_or_replace_dataset(
        raw_data_group, "display_mode", _dataset_display_mode, ascii=True
    )
    if file_path is not None:
        _create_or_replace_dataset(raw_data_group, "filePath", file_path, ascii=True)
    _create_or_replace_dataset(raw_data_group, "location", "FileSystem", ascii=True)
    if nickname is not None:
        _create_or_replace_dataset(raw_data_group, "nickname", nickname, ascii=True)
    _create_or_replace_dataset(raw_data_group, "normalizeDisplay", False)
    if shape is not None:
        _create_or_replace_dataset(
            raw_data_group, "shape", data=np.array(shape, dtype=np.int64)
        )


def _get_hdf5_file(hdf5_path: Union[str, PathLike]) -> Optional[Path]:
    hdf5_file = Path(hdf5_path)
    while hdf5_file is not None and hdf5_file.suffix != ".h5":
        hdf5_file = hdf5_file.parent
    return hdf5_file


def _str_decode(obj: Union[str, bytes]) -> Tuple[str, bool]:
    if isinstance(obj, bytes):
        return obj.decode("ascii"), True
    return obj, False


def _str_encode(obj: str, ascii: bool = False) -> Union[str, bytes]:
    if ascii and isinstance(obj, str):
        return obj.encode("ascii")
    return obj


def _create_or_replace_dataset(
    parent: Union[h5py.File, h5py.Group],
    key: str,
    data: Union[str, Any],
    ascii: bool = True,
) -> h5py.Dataset:
    old_dataset = parent.get(key)
    if old_dataset is not None:
        old_data = old_dataset[()]
        if isinstance(old_data, str):
            _, ascii = _str_decode(old_data)
        del parent[key]
    if isinstance(data, str):
        data = _str_encode(data, ascii=ascii)
    return parent.create_dataset(key, data=data)



import click

from ..._cli.utils import OrderedClickGroup
from .cellprofiler import cellprofiler_cmd_group
from .intensities import intensities_cmd
from .neighbors import neighbors_cmd
from .regionprops import regionprops_cmd


@click.group(
    name="measure",
    cls=OrderedClickGroup,
    help="Extract object data from segmented images",
)
def measure_cmd_group():
    pass


measure_cmd_group.add_command(intensities_cmd)
measure_cmd_group.add_command(regionprops_cmd)
measure_cmd_group.add_command(neighbors_cmd)
measure_cmd_group.add_command(cellprofiler_cmd_group)



import logging
from os import PathLike
from pathlib import Path
from typing import Generator, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from anndata import AnnData
from scipy.sparse import csr_matrix

from .. import io

logger = logging.getLogger(__name__)


def try_convert_to_dataframe_from_disk(
    *data_file_lists,
) -> Generator[Tuple[str, Tuple[Path, ...], pd.DataFrame], None, None]:
    for data_files in zip(*data_file_lists):
        data_files = tuple(Path(data_file) for data_file in data_files)
        img_file_name = io._as_path_with_suffix(data_files[0], ".tiff").name
        try:
            df = io.read_data(data_files[0])
            for data_file in data_files[1:]:
                df = pd.merge(
                    df,
                    io.read_data(data_file),
                    left_index=True,
                    right_index=True,
                )
            yield img_file_name, data_files, df
            del df
        except:
            logger.exception(
                f"Error creating DataFrame for image {img_file_name}; " "skipping image"
            )


def try_convert_to_anndata_from_disk(
    intensity_files: Sequence[Union[str, PathLike]],
    *data_file_lists,
    neighbors_files: Optional[Sequence[Union[str, PathLike]]] = None,
    # panel: Optional[pd.DataFrame] = None,
    # image_info: Optional[pd.DataFrame] = None,
) -> Generator[Tuple[str, Path, Tuple[Path, ...], Optional[Path], AnnData], None, None]:
    # if panel is not None:
    #     panel = panel.set_index("name", drop=False, verify_integrity=True)
    # if image_info is not None:
    #     image_info = image_info.set_index(
    #         "image", drop=False, verify_integrity=True
    #     )
    for i, intensity_file in enumerate(intensity_files):
        intensity_file = Path(intensity_file)
        data_files = tuple(Path(dfl[i]) for dfl in data_file_lists)
        neighbors_file = None
        if neighbors_files is not None:
            neighbors_file = Path(neighbors_files[i])
        img_file_name = io._as_path_with_suffix(intensity_file, ".tiff").name
        try:
            x = io.read_data(intensity_file)
            obs = None
            if len(data_files) > 0:
                obs = io.read_data(data_files[0])
                for data_file in data_files[1:]:
                    obs = pd.merge(
                        obs,
                        io.read_data(data_file),
                        left_index=True,
                        right_index=True,
                    )
                obs = obs.loc[x.index, :]
            # if image_info is not None:
            #     image_obs = (
            #         pd.concat(
            #             [image_info.loc[img_file_name, :]] * len(x.index),
            #             axis=1,
            #         )
            #         .transpose()
            #         .astype(image_info.dtypes.to_dict())
            #     )
            #     image_obs.index = x.index
            #     image_obs.columns = "image_" + image_obs.columns
            #     image_obs.rename(
            #         columns={"image_image": "image"}, inplace=True
            #     )
            #     if obs is not None:
            #         obs = pd.merge(
            #             obs,
            #             image_obs,
            #             how="inner",  # preserves order of left keys
            #             left_index=True,
            #             right_index=True,
            #         )
            #     else:
            #         obs = image_obs
            var = None
            # if panel is not None:
            #     var = panel.loc[x.columns, :].copy()
            if obs is not None:
                obs.index = [f"Object {object_id}" for object_id in x.index]
            if var is not None:
                var.index = x.columns.astype(str).tolist()
            adata = AnnData(X=x.values, obs=obs, var=var, dtype=np.float32)
            if neighbors_file is not None:
                neighbors = io.read_neighbors(neighbors_file)
                row_ind = [x.index.get_loc(a) for a in neighbors["Object"]]
                col_ind = [x.index.get_loc(b) for b in neighbors["Neighbor"]]
                adata.obsp["adj"] = csr_matrix(
                    ([True] * len(neighbors.index), (row_ind, col_ind)),
                    shape=(adata.n_obs, adata.n_obs),
                    dtype=np.uint8,
                )
                if neighbors["Distance"].notna().any():
                    adata.obsp["dist"] = csr_matrix(
                        (neighbors["Distance"].values, (row_ind, col_ind)),
                        shape=(adata.n_obs, adata.n_obs),
                        dtype=np.float32,
                    )
                del neighbors
            yield (
                img_file_name,
                intensity_file,
                data_files,
                neighbors_file,
                adata,
            )
            del x, obs, var, adata
        except:
            logger.exception(
                f"Error creating AnnData object for image {img_file_name}; "
                "skipping image"
            )



from pathlib import Path

import click
import click_log

from ... import io
from ..._cli.utils import catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from ..neighbors import NeighborhoodType, try_measure_neighbors_from_disk

_neighborhood_types = {
    "centroids": NeighborhoodType.CENTROID_DISTANCE,
    "borders": NeighborhoodType.EUCLIDEAN_BORDER_DISTANCE,
    "expansion": NeighborhoodType.EUCLIDEAN_PIXEL_EXPANSION,
}


@click.command(name="neighbors", help="Measure object neighbors")
@click.option(
    "--masks",
    "mask_dir",
    type=click.Path(exists=True, file_okay=False),
    default="masks",
    show_default=True,
    help="Path to the mask directory",
)
@click.option(
    "--type",
    "neighborhood_type_name",
    type=click.Choice(list(_neighborhood_types.keys()), case_sensitive=True),
    required=True,
    help="Neighborhood type",
)
@click.option(
    "--metric",
    "metric",
    type=click.STRING,
    default="euclidean",
    help="Spatial distance metric (see scipy's pdist)",
)
@click.option(
    "--dmax",
    "dmax",
    type=click.FLOAT,
    help="Maximum spatial distance between objects",
)
@click.option(
    "--kmax",
    "kmax",
    type=click.INT,
    help="Maximum number of neighbors per object",
)
@click.option(
    "--mmap/--no-mmap",
    "mmap",
    default=False,
    show_default=True,
    help="Use memory mapping for reading images/masks",
)
@click.option(
    "-o",
    "neighbors_dir",
    type=click.Path(file_okay=False),
    default="neighbors",
    show_default=True,
    help="Path to the object neighbors output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def neighbors_cmd(
    mask_dir, neighborhood_type_name, metric, dmax, kmax, mmap, neighbors_dir
):
    mask_files = io.list_mask_files(mask_dir)
    Path(neighbors_dir).mkdir(exist_ok=True)
    for mask_file, neighbors in try_measure_neighbors_from_disk(
        mask_files,
        _neighborhood_types[neighborhood_type_name],
        metric=metric,
        dmax=dmax,
        kmax=kmax,
        mmap=mmap,
    ):
        neighbors_file = io._as_path_with_suffix(
            Path(neighbors_dir) / Path(mask_file).name, ".csv"
        )
        io.write_neighbors(neighbors, neighbors_file)
        logger.info(neighbors_file)
        del neighbors



from pathlib import Path

from steinbock import io
from steinbock.export import data


class TestDataExport:
    def test_try_convert_to_dataframe_from_disk(
        self, imc_test_data_steinbock_path: Path
    ):
        intensities_files = io.list_data_files(
            imc_test_data_steinbock_path / "intensities"
        )
        regionprops_files = io.list_data_files(
            imc_test_data_steinbock_path / "regionprops",
            base_files=intensities_files,
        )
        gen = data.try_convert_to_dataframe_from_disk(
            intensities_files, regionprops_files
        )
        for img_file_name, data_files, df in gen:
            pass  # TODO

    def test_try_convert_to_anndata_from_disk(self, imc_test_data_steinbock_path: Path):
        intensities_files = io.list_data_files(
            imc_test_data_steinbock_path / "intensities"
        )
        regionprops_files = io.list_data_files(
            imc_test_data_steinbock_path / "regionprops",
            base_files=intensities_files,
        )
        neighbors_files = io.list_neighbors_files(
            imc_test_data_steinbock_path / "neighbors",
            base_files=intensities_files,
        )
        gen = data.try_convert_to_anndata_from_disk(
            intensities_files,
            regionprops_files,
            neighbors_files=neighbors_files,
        )
        for (
            img_file_name,
            intensities_file,
            (regionprops_file,),
            neighbors_file,
            adata,
        ) in gen:
            pass  # TODO



import logging
from collections import OrderedDict
from functools import partial, wraps
from typing import Dict

import click
from click.core import Command

from .._steinbock import SteinbockException


logger = logging.getLogger(__name__.rpartition(".")[0].rpartition(".")[0])


class SteinbockCLIException(SteinbockException):
    pass


class OrderedClickGroup(click.Group):
    def __init__(self, *args, commands=None, **kwargs) -> None:
        super(OrderedClickGroup, self).__init__(*args, **kwargs)
        self.commands = commands or OrderedDict()

    def list_commands(self, ctx) -> Dict[str, Command]:
        return self.commands


def catch_exception(func=None, *, handle=SteinbockException):
    if not func:
        return partial(catch_exception, handle=handle)

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except handle as e:
            raise click.ClickException(e)

    return wrapper



from pathlib import Path

import numpy as np
import pytest
from steinbock import io
from steinbock.segmentation import deepcell
from steinbock.segmentation.deepcell import Application

keras_models_dir = "/opt/keras/models"


@pytest.mark.skipif(not deepcell.deepcell_available, reason="DeepCell is not available")
class TestDeepcellSegmentation:
    @pytest.mark.skip(reason="Test would take too long")
    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    def test_try_segment_objects_mesmer(self, imc_test_data_steinbock_path: Path):
        from tensorflow.keras.models import load_model  # type: ignore

        img_files = io.list_image_files(imc_test_data_steinbock_path / "img")
        model = None
        model_path = Path(keras_models_dir) / "MultiplexSegmentation"
        if model_path.exists():
            model = load_model(model_path, compile=False)
        channel_groups = np.array([np.nan, 2, np.nan, np.nan, 1])
        deepcell.try_segment_objects(
            img_files,
            Application.MESMER,
            model=model,
            channelwise_minmax=True,
            channel_groups=channel_groups,
        )  # TODO



# from pathlib import Path

# from steinbock import io
# from steinbock.utils import mosaics


# class TestMosaicsUtils:
#     def test_try_extract_tiles_from_disk(
#         self, imc_test_data_steinbock_path: Path
#     ):
#         img_files = io.list_image_files(imc_test_data_steinbock_path / "img")
#         gen = mosaics.try_extract_tiles_from_disk(img_files, 50)
#         for img_file, tile_info, tile in gen:
#             pass  # TODO

#     def test_try_stitch_tiles_from_disk(
#         self, imc_test_data_steinbock_path: Path, tmp_path: Path
#     ):
#         img_files = io.list_image_files(imc_test_data_steinbock_path / "img")
#         img_file_stems = []
#         img_tile_files = {}
#         img_tile_infos = {}
#         gen = mosaics.try_extract_tiles_from_disk(img_files, 50)
#         for img_file, tile_info, tile in gen:
#             tile_file = tmp_path / (
#                 f"{tile_info.img_file_stem}_tx{tile_info.x}_ty{tile_info.y}"
#                 f"_tw{tile_info.width}_th{tile_info.height}.tiff"
#             )
#             if tile_info.img_file_stem not in img_file_stems:
#                 img_file_stems.append(tile_info.img_file_stem)
#                 img_tile_files[tile_info.img_file_stem] = []
#                 img_tile_infos[tile_info.img_file_stem] = []
#             img_tile_files[tile_info.img_file_stem].append(tile_file)
#             img_tile_infos[tile_info.img_file_stem].append(tile_info)
#             io.write_image(tile, tile_file, ignore_dtype=True)
#             del tile
#         gen = mosaics.try_stitch_tiles_from_disk(
#             img_file_stems, img_tile_files, img_tile_infos
#         )
#         for img_file_stem, img in gen:
#             pass  # TODO



import logging
from enum import Enum
from functools import partial
from importlib.util import find_spec
from os import PathLike
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np

from .. import io
from ._segmentation import SteinbockSegmentationException

if TYPE_CHECKING:
    from tensorflow.keras.models import Model  # type: ignore


logger = logging.getLogger(__name__)
deepcell_available = find_spec("deepcell") is not None


class SteinbockDeepcellSegmentationException(SteinbockSegmentationException):
    pass


def _mesmer_application(model=None):
    from deepcell.applications import Mesmer

    app = Mesmer(model=model)

    def predict(
        img: np.ndarray,
        *,
        pixel_size_um: Optional[float] = None,
        segmentation_type: Optional[str] = None,
        preprocess_kwargs: Optional[Mapping[str, Any]] = None,
        postprocess_kwargs: Optional[Mapping[str, Any]] = None,
    ) -> np.ndarray:
        assert img.ndim == 3
        if pixel_size_um is None:
            raise SteinbockDeepcellSegmentationException("Unknown pixel size")
        if segmentation_type is None:
            raise SteinbockDeepcellSegmentationException("Unknown segmentation type")
        mask = app.predict(
            np.expand_dims(np.moveaxis(img, 0, -1), 0),
            batch_size=1,
            image_mpp=pixel_size_um,
            compartment=segmentation_type,
            preprocess_kwargs=preprocess_kwargs or {},
            postprocess_kwargs_whole_cell=postprocess_kwargs or {},
            postprocess_kwargs_nuclear=postprocess_kwargs or {},
        )[0, :, :, 0]
        assert mask.shape == img.shape[1:]
        return mask

    return app, predict


class Application(Enum):
    """"""

    MESMER = partial(_mesmer_application)


def try_segment_objects(
    img_files: Sequence[Union[str, PathLike]],
    application: Application,
    model: Optional["Model"] = None,
    channelwise_minmax: bool = False,
    channelwise_zscore: bool = False,
    channel_groups: Optional[np.ndarray] = None,
    aggr_func: Callable[[np.ndarray], np.ndarray] = np.mean,
    **predict_kwargs,
) -> Generator[Tuple[Path, np.ndarray], None, None]:
    app, predict = application.value(model=model)
    for img_file in img_files:
        try:
            img = io.read_image(img_file)
            if channelwise_minmax:
                channel_mins = np.nanmin(img, axis=(1, 2))
                channel_maxs = np.nanmax(img, axis=(1, 2))
                channel_ranges = channel_maxs - channel_mins
                img -= channel_mins[:, np.newaxis, np.newaxis]
                img[channel_ranges > 0] /= channel_ranges[
                    channel_ranges > 0, np.newaxis, np.newaxis
                ]
            if channelwise_zscore:
                channel_means = np.nanmean(img, axis=(1, 2))
                channel_stds = np.nanstd(img, axis=(1, 2))
                img -= channel_means[:, np.newaxis, np.newaxis]
                img[channel_stds > 0] /= channel_stds[
                    channel_stds > 0, np.newaxis, np.newaxis
                ]
            if channel_groups is not None:
                img = np.stack(
                    [
                        aggr_func(img[channel_groups == channel_group], axis=0)
                        for channel_group in np.unique(channel_groups)
                        if not np.isnan(channel_group)
                    ]
                )
            mask = predict(img, **predict_kwargs)
            yield Path(img_file), mask
            del img, mask
        except:
            logger.exception(f"Error segmenting objects in {img_file}")



from pathlib import Path

import click
import click_log
import numpy as np

from ... import io
from ..._cli.utils import catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import deepcell

if deepcell.deepcell_available:
    import yaml
else:
    yaml = None

deepcell_cli_available = deepcell.deepcell_available

_applications = {
    "mesmer": deepcell.Application.MESMER,
}


@click.command(name="deepcell", help="Run a object segmentation batch using DeepCell")
@click.option(
    "--app",
    "application_name",
    type=click.Choice(list(_applications.keys()), case_sensitive=True),
    show_choices=True,
    default="mesmer",
    show_default=True,
    help="DeepCell application name",
)
@click.option(
    "--model",
    "model_path_or_name",
    type=click.STRING,
    default="MultiplexSegmentation",
    show_default=True,
    help="Path/name of custom Keras model",
)
@click.option(
    "--modeldir",
    "keras_model_dir",
    type=click.Path(file_okay=False),
    default="/opt/keras/models",
    show_default=True,
    help="Path to Keras model directory",
)
@click.option(
    "--img",
    "img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image directory",
)
@click.option(
    "--minmax/--no-minmax",
    "channelwise_minmax",
    default=False,
    show_default=True,
    help="Channel-wise min-max normalization",
)
@click.option(
    "--zscore/--no-zscore",
    "channelwise_zscore",
    default=False,
    show_default=True,
    help="Channel-wise z-score normalization",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(exists=True, dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel file",
)
@click.option(
    "--aggr",
    "aggr_func_name",
    type=click.STRING,
    default="mean",
    show_default=True,
    help="Numpy function for aggregating channel pixels",
)
@click.option(
    "--pixelsize",
    "pixel_size_um",
    type=click.FLOAT,
    default=1.0,
    show_default=True,
    help="[Mesmer] Pixel size in micrometers",
)
@click.option(
    "--type",
    "segmentation_type",
    type=click.Choice(["whole-cell", "nuclear"]),
    default="whole-cell",
    show_default=True,
    show_choices=True,
    help="[Mesmer] Segmentation type",
)
@click.option(
    "--preprocess",
    "preprocess_file",
    type=click.Path(exists=True, dir_okay=False),
    help="[Mesmer] Preprocessing parameters (YAML file)",
)
@click.option(
    "--postprocess",
    "postprocess_file",
    type=click.Path(exists=True, dir_okay=False),
    help="[Mesmer] Postprocessing parameters (YAML file)",
)
@click.option(
    "-o",
    "mask_dir",
    type=click.Path(file_okay=False),
    default="masks",
    show_default=True,
    help="Path to the mask output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def deepcell_cmd(
    application_name,
    model_path_or_name,
    keras_model_dir,
    img_dir,
    channelwise_minmax,
    channelwise_zscore,
    panel_file,
    aggr_func_name,
    pixel_size_um,
    segmentation_type,
    preprocess_file,
    postprocess_file,
    mask_dir,
):
    channel_groups = None
    if Path(panel_file).exists():
        panel = io.read_panel(panel_file)
        if "deepcell" in panel and panel["deepcell"].notna().any():
            channel_groups = panel["deepcell"].values
    aggr_func = getattr(np, aggr_func_name)
    img_files = io.list_image_files(img_dir)
    model = None
    if model_path_or_name is not None:
        from tensorflow.keras.models import load_model  # type: ignore

        if Path(model_path_or_name).exists():
            model = load_model(model_path_or_name, compile=False)
        elif Path(keras_model_dir).joinpath(model_path_or_name).exists():
            model = load_model(
                Path(keras_model_dir).joinpath(model_path_or_name),
                compile=False,
            )
    preprocess_kwargs = None
    if preprocess_file is not None:
        with Path(preprocess_file).open() as f:
            preprocess_kwargs = yaml.load(f, yaml.Loader)
    postprocess_kwargs = None
    if postprocess_file is not None:
        with Path(postprocess_file).open() as f:
            postprocess_kwargs = yaml.load(f, yaml.Loader)
    Path(mask_dir).mkdir(exist_ok=True)
    for img_file, mask in deepcell.try_segment_objects(
        img_files,
        _applications[application_name],
        model=model,
        channelwise_minmax=channelwise_minmax,
        channelwise_zscore=channelwise_zscore,
        channel_groups=channel_groups,
        aggr_func=aggr_func,
        pixel_size_um=pixel_size_um,
        segmentation_type=segmentation_type,
        preprocess_kwargs=preprocess_kwargs,
        postprocess_kwargs=postprocess_kwargs,
    ):
        mask_file = io._as_path_with_suffix(Path(mask_dir) / img_file.name, ".tiff")
        io.write_mask(mask, mask_file)
        logger.info(mask_file)



from pathlib import Path

import numpy as np
from steinbock import io
from steinbock.measurement import regionprops


class TestRegionpropsMeasurement:
    def test_measure_regionprops(self):
        img = np.array(
            [
                [
                    [0.5, 1.5, 0.1],
                    [0.1, 0.2, 0.3],
                    [0.1, 6.5, 3.5],
                ],
                [
                    [20, 30, 1],
                    [1, 2, 3],
                    [1, 100, 200],
                ],
            ],
            dtype=io.img_dtype,
        )
        mask = np.array(
            [
                [1, 1, 0],
                [0, 0, 0],
                [0, 2, 2],
            ],
            dtype=io.mask_dtype,
        )
        regionprops.measure_regionprops(img, mask, ["area"])  # TODO

    def test_try_measure_regionprops_from_disk(
        self, imc_test_data_steinbock_path: Path
    ):
        img_files = io.list_image_files(imc_test_data_steinbock_path / "img")
        mask_files = io.list_mask_files(
            imc_test_data_steinbock_path / "masks", base_files=img_files
        )
        gen = regionprops.try_measure_regionprops_from_disk(
            img_files, mask_files, ["area"]
        )
        for img_file, mask_file, df in gen:
            pass  # TODO



from pathlib import Path

import click
import click_log

from ... import io
from ..._cli.utils import catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from ..intensities import IntensityAggregation, try_measure_intensities_from_disk

_intensity_aggregations = {
    "sum": IntensityAggregation.SUM,
    "min": IntensityAggregation.MIN,
    "max": IntensityAggregation.MAX,
    "mean": IntensityAggregation.MEAN,
    "median": IntensityAggregation.MEDIAN,
    "std": IntensityAggregation.STD,
    "var": IntensityAggregation.VAR,
}


@click.command(name="intensities", help="Measure object intensities")
@click.option(
    "--img",
    "img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image directory",
)
@click.option(
    "--masks",
    "mask_dir",
    type=click.Path(exists=True, file_okay=False),
    default="masks",
    show_default=True,
    help="Path to the mask directory",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(exists=True, dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel file",
)
@click.option(
    "--aggr",
    "intensity_aggregation_name",
    type=click.Choice(list(_intensity_aggregations.keys()), case_sensitive=True),
    default="mean",
    show_default=True,
    help="Function for aggregating cell pixels",
)
@click.option(
    "--mmap/--no-mmap",
    "mmap",
    default=False,
    show_default=True,
    help="Use memory mapping for reading images/masks",
)
@click.option(
    "-o",
    "intensities_dir",
    type=click.Path(file_okay=False),
    default="intensities",
    show_default=True,
    help="Path to the object intensities output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def intensities_cmd(
    img_dir,
    mask_dir,
    panel_file,
    intensity_aggregation_name,
    mmap,
    intensities_dir,
):
    panel = io.read_panel(panel_file)
    channel_names = panel["name"].tolist()
    img_files = io.list_image_files(img_dir)
    mask_files = io.list_mask_files(mask_dir, base_files=img_files)
    Path(intensities_dir).mkdir(exist_ok=True)
    for img_file, mask_file, intensities in try_measure_intensities_from_disk(
        img_files,
        mask_files,
        channel_names,
        _intensity_aggregations[intensity_aggregation_name],
        mmap=mmap,
    ):
        intensities_file = io._as_path_with_suffix(
            Path(intensities_dir) / img_file.name, ".csv"
        )
        io.write_data(intensities, intensities_file)
        logger.info(intensities_file)
        del intensities



from contextlib import nullcontext
from os import PathLike
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Union
from zipfile import ZipFile

import click
import click_log
import pandas as pd

from ... import io
from ..._cli.utils import (
    OrderedClickGroup,
    SteinbockCLIException,
    catch_exception,
    logger,
)
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import imc

imc_cli_available = imc.imc_available


@click.group(
    name="imc",
    cls=OrderedClickGroup,
    help="Preprocess Imaging Mass Cytometry (IMC) data",
)
def imc_cmd_group():
    pass


@imc_cmd_group.command(name="panel", help="Create a panel from IMC panel/image data")
@click.option(
    "--imcpanel",
    "imc_panel_file",
    type=click.Path(dir_okay=False),
    default=str(Path("raw", "panel.csv")),
    show_default=True,
    help="Path to the IMC panel file",
)
@click.option(
    "--channelcol",
    "imc_panel_channel_col",
    type=click.STRING,
    default="Metal Tag",
    show_default=True,
    help="IMC panel column containing the channel (required)",
)
@click.option(
    "--namecol",
    "imc_panel_name_col",
    type=click.STRING,
    default="Target",
    show_default=True,
    help="IMC panel column containing the channel name (required)",
)
@click.option(
    "--keepcol",
    "imc_panel_keep_col",
    type=click.STRING,
    default="full",
    show_default=True,
    help="IMC panel column indicating whether to keep the channel (0/1)",
)
@click.option(
    "--ilastikcol",
    "imc_panel_ilastik_col",
    type=click.STRING,
    default="ilastik",
    show_default=True,
    help="IMC panel column indicating Ilastik usage (0/1)",
)
@click.option(
    "--mcd",
    "mcd_dir",
    type=click.Path(file_okay=False),
    default="raw",
    show_default=True,
    help="Path to the IMC .mcd file directory",
)
@click.option(
    "--txt",
    "txt_dir",
    type=click.Path(file_okay=False),
    default="raw",
    show_default=True,
    help="Path to the IMC .txt file directory",
)
@click.option(
    "-o",
    "panel_file",
    type=click.Path(dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel output file",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def panel_cmd(
    imc_panel_file,
    imc_panel_channel_col,
    imc_panel_name_col,
    imc_panel_keep_col,
    imc_panel_ilastik_col,
    mcd_dir,
    txt_dir,
    panel_file,
):
    panel = None
    if panel is None and Path(imc_panel_file).exists():
        panel = imc.create_panel_from_imc_panel(
            imc_panel_file,
            imc_panel_channel_col=imc_panel_channel_col,
            imc_panel_name_col=imc_panel_name_col,
            imc_panel_keep_col=imc_panel_keep_col,
            imc_panel_ilastik_col=imc_panel_ilastik_col,
        )
    if panel is None and Path(mcd_dir).exists():
        mcd_files = _collect_mcd_files(mcd_dir)
        if len(mcd_files) > 0:
            panel = imc.create_panel_from_mcd_files(mcd_files)
    if panel is None and Path(txt_dir).exists():
        txt_files = _collect_txt_files(txt_dir)
        if len(txt_files) > 0:
            panel = imc.create_panel_from_txt_files(txt_files)
    if panel is None:
        raise SteinbockCLIException("No panel/.mcd/.txt file found")
    io.write_panel(panel, panel_file)
    logger.info(panel_file)


@imc_cmd_group.command(
    name="images", help="Extract Imaging Mass Cytometry (IMC) images"
)
@click.option(
    "--mcd",
    "mcd_dir",
    type=click.Path(exists=True, file_okay=False),
    default="raw",
    show_default=True,
    help="Path to the IMC .mcd file directory",
)
@click.option(
    "--txt",
    "txt_dir",
    type=click.Path(exists=True, file_okay=False),
    default="raw",
    show_default=True,
    help="Path to the IMC .txt file directory",
)
@click.option(
    "--unzip/--no-unzip",
    "unzip",
    default=True,
    show_default=True,
    help="Unzip .mcd/.txt files from .zip archives",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the steinbock panel file",
)
@click.option(
    "--hpf",
    "hpf",
    type=click.FLOAT,
    help="Hot pixel filter (specify delta threshold)",
)
@click.option(
    "--imgout",
    "img_dir",
    type=click.Path(file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image output directory",
)
@click.option(
    "--infoout",
    "image_info_file",
    type=click.Path(dir_okay=False),
    default="images.csv",
    show_default=True,
    help="Path to the image information output file",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def images_cmd(mcd_dir, txt_dir, unzip, panel_file, hpf, img_dir, image_info_file):
    channel_names = None
    if Path(panel_file).exists():
        panel = io.read_panel(panel_file)
        if "channel" in panel:
            channel_names = panel["channel"].tolist()
    image_info_data = []
    with (TemporaryDirectory() if unzip else nullcontext()) as temp_dir:
        mcd_files = _collect_mcd_files(mcd_dir, unzip_dir=temp_dir)
        txt_files = _collect_txt_files(txt_dir, unzip_dir=temp_dir)
        Path(img_dir).mkdir(exist_ok=True)
        mcd_txt_files = {}
        num_dupl = 0
        for (
            mcd_txt_file,
            acquisition,
            img,
            recovery_file,
            recovered,
        ) in imc.try_preprocess_images_from_disk(
            mcd_files, txt_files, channel_names=channel_names, hpf=hpf
        ):
            img_file_stem = Path(mcd_txt_file).stem
            if acquisition is not None:
                img_file_stem += f"_{acquisition.id:03d}"
            if img_file_stem in mcd_txt_files:
                num_dupl += 1
                first_mcd_txt_file = mcd_txt_files[img_file_stem][0]
                img_file_stem = f"DUPLICATE{num_dupl:03d}_{img_file_stem}"
                logger.warning(
                    f"File {mcd_txt_file} is a duplicate of {first_mcd_txt_file}, "
                    f"saving as {img_file_stem}"
                )
                mcd_txt_files[img_file_stem].append(mcd_txt_file)
            else:
                mcd_txt_files[img_file_stem] = [mcd_txt_file]
            img_file = Path(img_dir) / f"{img_file_stem}.tiff"
            io.write_image(img, img_file)
            recovery_file_name = None
            if recovery_file is not None:
                recovery_file_name = recovery_file.name
            image_info_row = {
                "image": img_file.name,
                "width_px": img.shape[2],
                "height_px": img.shape[1],
                "num_channels": img.shape[0],
                "source_file": mcd_txt_file.name,
                "recovery_file": recovery_file_name,
                "recovered": recovered,
            }
            if acquisition is not None:
                image_info_row.update(
                    {
                        "acquisition_id": acquisition.id,
                        "acquisition_description": acquisition.description,
                        "acquisition_start_x_um": (acquisition.roi_points_um[0][0]),
                        "acquisition_start_y_um": (acquisition.roi_points_um[0][1]),
                        "acquisition_end_x_um": (acquisition.roi_points_um[2][0]),
                        "acquisition_end_y_um": (acquisition.roi_points_um[2][1]),
                        "acquisition_width_um": acquisition.width_um,
                        "acquisition_height_um": acquisition.height_um,
                    }
                )
            image_info_data.append(image_info_row)
            logger.info(img_file)
            del img
    image_info = pd.DataFrame(data=image_info_data)
    io.write_image_info(image_info, image_info_file)
    logger.info(image_info_file)


def _extract_zips(
    path: Union[str, PathLike], suffix: str, dest: Union[str, PathLike]
) -> List[Path]:
    extracted_files = []
    for zip_file_path in Path(path).rglob("*.zip"):
        with ZipFile(zip_file_path) as zip_file:
            zip_infos = sorted(zip_file.infolist(), key=lambda x: x.filename)
            for zip_info in zip_infos:
                if not zip_info.is_dir() and zip_info.filename.endswith(suffix):
                    extracted_file = zip_file.extract(zip_info, path=dest)
                    extracted_files.append(Path(extracted_file))
    return extracted_files


def _collect_mcd_files(
    mcd_dir: Union[str, PathLike], unzip_dir: bool = None
) -> List[Path]:
    mcd_files = imc.list_mcd_files(mcd_dir)
    if unzip_dir is not None:
        mcd_files += _extract_zips(mcd_dir, ".mcd", unzip_dir)
    return mcd_files


def _collect_txt_files(
    txt_dir: Union[str, PathLike], unzip_dir: bool = None
) -> List[Path]:
    txt_files = imc.list_txt_files(txt_dir)
    if unzip_dir is not None:
        txt_files += _extract_zips(txt_dir, ".txt", unzip_dir)
    return txt_files



import re
from pathlib import Path

import click
import click_log
import numpy as np
import tifffile
import xtiff

from ... import io
from ..._cli.utils import OrderedClickGroup, catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .data import anndata_cmd, csv_cmd, fcs_cmd
from .graphs import graphs_cmd


@click.group(
    name="export",
    cls=OrderedClickGroup,
    help="Export data to third-party formats",
)
def export_cmd_group():
    pass


@export_cmd_group.command(name="ome", help="Export images as OME-TIFF")
@click.option(
    "--img",
    "img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image directory",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(exists=True, dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel file",
)
@click.option(
    "-o",
    "ome_dir",
    type=click.Path(file_okay=False),
    default="ome",
    show_default=True,
    help="Path to the OME-TIFF export directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def ome_cmd(img_dir, panel_file, ome_dir):
    panel = io.read_panel(panel_file)
    channel_names = [
        f"{channel_id}_{channel_name}"
        for channel_id, channel_name in zip(
            panel["channel"].values, panel["name"].values
        )
    ]
    Path(ome_dir).mkdir(exist_ok=True)
    for img_file in io.list_image_files(img_dir):
        img = io.read_image(img_file, native_dtype=True)
        ome_file = io._as_path_with_suffix(Path(ome_dir) / img_file.name, ".ome.tiff")
        xtiff.to_tiff(
            io._to_dtype(img, np.float32),
            ome_file,
            channel_names=channel_names,
        )
        logger.info(ome_file)
        del img


@export_cmd_group.command(name="histocat", help="Export images to histoCAT for MATLAB")
@click.option(
    "--img",
    "img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image directory",
)
@click.option(
    "--masks",
    "mask_dir",
    type=click.Path(file_okay=False),
    default="masks",
    show_default=True,
    help="Path to the mask directory",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(exists=True, dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel file",
)
@click.option(
    "-o",
    "histocat_dir",
    type=click.Path(file_okay=False),
    default="histocat",
    show_default=True,
    help="Path to the histoCAT export directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def histocat_cmd(img_dir, mask_dir, panel_file, histocat_dir):
    panel = io.read_panel(panel_file)
    channel_names = [
        f"{channel_id}_{channel_name}"
        for channel_id, channel_name in zip(
            panel["channel"].values, panel["name"].values
        )
    ]
    img_files = io.list_image_files(img_dir)
    mask_files = None
    if Path(mask_dir).exists():
        mask_files = io.list_mask_files(mask_dir, base_files=img_files)
    Path(histocat_dir).mkdir(exist_ok=True)
    for i, img_file in enumerate(img_files):
        img = io.read_image(img_file, native_dtype=True)
        histocat_img_dir = Path(histocat_dir) / img_file.stem
        histocat_img_dir.mkdir(exist_ok=True)
        for channel_name, channel_img in zip(channel_names, img):
            channel_name = re.sub(r"\s*/\s*", "_", channel_name)
            channel_name = re.sub(r"\s", "_", channel_name)
            histocat_img_file = histocat_img_dir / f"{channel_name}.tiff"
            tifffile.imwrite(
                histocat_img_file,
                data=io._to_dtype(channel_img, np.float32)[
                    np.newaxis, np.newaxis, np.newaxis, :, :, np.newaxis
                ],
                imagej=True,
            )
            logger.info(histocat_img_file)
        mask = None
        if mask_files is not None:
            mask = io.read_mask(mask_files[i], native_dtype=True)
            histocat_mask_file = histocat_img_dir / f"{img_file.stem}_mask.tiff"
            tifffile.imwrite(
                histocat_mask_file,
                data=io._to_dtype(mask, np.uint16)[
                    np.newaxis, np.newaxis, np.newaxis, :, :, np.newaxis
                ],
                imagej=True,
            )
            logger.info(histocat_mask_file)
        del img, mask


export_cmd_group.add_command(csv_cmd)
export_cmd_group.add_command(fcs_cmd)
export_cmd_group.add_command(anndata_cmd)
export_cmd_group.add_command(graphs_cmd)



import click

from .._version import version as steinbock_version
from ..classification._cli import classify_cmd_group
from ..export._cli import export_cmd_group
from ..measurement._cli import measure_cmd_group
from ..preprocessing._cli import preprocess_cmd_group
from ..segmentation._cli import segment_cmd_group
from ..utils._cli import utils_cmd_group
from .apps import apps_cmd_group
from .utils import OrderedClickGroup


@click.group(name="steinbock", cls=OrderedClickGroup)
@click.version_option(steinbock_version)
def steinbock_cmd_group():
    pass


steinbock_cmd_group.add_command(preprocess_cmd_group)
steinbock_cmd_group.add_command(classify_cmd_group)
steinbock_cmd_group.add_command(segment_cmd_group)
steinbock_cmd_group.add_command(measure_cmd_group)
steinbock_cmd_group.add_command(export_cmd_group)
steinbock_cmd_group.add_command(utils_cmd_group)
steinbock_cmd_group.add_command(apps_cmd_group)



import logging
from os import PathLike
from pathlib import Path
from typing import Generator, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from .. import io

logger = logging.getLogger(__name__)


def match_masks(mask1: np.ndarray, mask2: np.ndarray) -> pd.DataFrame:
    nz1 = mask1 != 0
    nz2 = mask2 != 0
    object_ids1 = []
    object_ids2 = []
    for object_id1 in np.unique(mask1[nz1]):
        for object_id2 in np.unique(mask2[nz2 & (mask1 == object_id1)]):
            object_ids1.append(object_id1)
            object_ids2.append(object_id2)
    for object_id2 in np.unique(mask2[nz2]):
        for object_id1 in np.unique(mask1[nz1 & (mask2 == object_id2)]):
            object_ids1.append(object_id1)
            object_ids2.append(object_id2)
    df = pd.DataFrame(data={"Object1": object_ids1, "Object2": object_ids2})
    df.drop_duplicates(inplace=True, ignore_index=True)
    return df


def try_match_masks_from_disk(
    mask_files1: Sequence[Union[str, PathLike]],
    mask_files2: Sequence[Union[str, PathLike]],
    mmap: bool = False,
) -> Generator[Tuple[Path, Path, pd.DataFrame], None, None]:
    for mask_file1, mask_file2 in zip(mask_files1, mask_files2):
        try:
            if mmap:
                mask1 = io.mmap_mask(mask_file1)
                mask2 = io.mmap_mask(mask_file2)
            else:
                mask1 = io.read_mask(mask_file1)
                mask2 = io.read_mask(mask_file2)
            df = match_masks(mask1, mask2)
            del mask1, mask2
            yield mask_file1, mask_file2, df
            del df
        except:
            logger.exception(f"Error matching masks {mask_file1, mask_file2}")



from pathlib import Path

import click
import click_log

from ... import io
from ..._cli.utils import catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import matching


@click.command(name="match", help="Match mask objects")
@click.argument("masks1", nargs=1, type=click.Path(exists=True, file_okay=False))
@click.argument("masks2", nargs=1, type=click.Path(exists=True, file_okay=False))
@click.option(
    "--mmap/--no-mmap",
    "mmap",
    default=False,
    show_default=True,
    help="Use memory mapping for reading images/masks",
)
@click.option(
    "-o",
    "csv_dir",
    type=click.Path(file_okay=False),
    required=True,
    help="Path to the object table CSV output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def match_cmd(masks1, masks2, mmap, csv_dir):
    if Path(masks1).is_file() and Path(masks2).is_file():
        mask_files1 = [Path(masks1)]
        mask_files2 = [Path(masks2)]
    elif Path(masks1).is_dir() and Path(masks2).is_dir():
        mask_files1 = io.list_mask_files(masks1)
        mask_files2 = io.list_mask_files(masks2, base_files=mask_files1)
    Path(csv_dir).mkdir(exist_ok=True)
    for mask_file1, mask_file2, df in matching.try_match_masks_from_disk(
        mask_files1, mask_files2, mmap=mmap
    ):
        csv_file = io._as_path_with_suffix(Path(csv_dir) / mask_file1.name, ".csv")
        df.columns = [Path(masks1).name, Path(masks2).name]
        df.to_csv(csv_file, index=False)
        logger.info(csv_file)
        del df



from pathlib import Path

from steinbock import io


class TestIO:
    def test_read_panel(self, imc_test_data_steinbock_path: Path):
        io.read_panel(imc_test_data_steinbock_path / "panel.csv")  # TODO

    def test_write_panel(self, imc_test_data_steinbock_path: Path):
        pass  # TODO

    def test_list_image_files(self, imc_test_data_steinbock_path: Path):
        io.list_image_files(imc_test_data_steinbock_path / "img")  # TODO

    def test_read_image(self, imc_test_data_steinbock_path: Path):
        io.read_image(
            imc_test_data_steinbock_path / "img" / "20210305_NE_mockData1_1.tiff"
        )  # TODO

    def test_write_image(self, imc_test_data_steinbock_path: Path):
        pass  # TODO

    def test_read_image_info(self, imc_test_data_steinbock_path: Path):
        io.read_image_info(imc_test_data_steinbock_path / "images.csv")  # TODO

    def test_write_image_info(self, imc_test_data_steinbock_path: Path):
        pass  # TODO

    def test_list_mask_files(self, imc_test_data_steinbock_path: Path):
        io.list_mask_files(imc_test_data_steinbock_path / "masks")  # TODO

    def test_read_mask(self, imc_test_data_steinbock_path: Path):
        io.read_mask(
            imc_test_data_steinbock_path / "masks" / "20210305_NE_mockData1_1.tiff"
        )  # TODO

    def test_write_mask(self, imc_test_data_steinbock_path: Path):
        pass  # TODO

    def test_list_data_files(self, imc_test_data_steinbock_path: Path):
        io.list_data_files(imc_test_data_steinbock_path / "intensities")  # TODO

    def test_read_data(self, imc_test_data_steinbock_path: Path):
        io.read_data(
            imc_test_data_steinbock_path / "intensities" / "20210305_NE_mockData1_1.csv"
        )  # TODO

    def test_write_data(self, imc_test_data_steinbock_path: Path):
        pass  # TODO

    def test_list_neighbors_files(self, imc_test_data_steinbock_path: Path):
        io.list_neighbors_files(imc_test_data_steinbock_path / "neighbors")  # TODO

    def test_read_neighbors(self, imc_test_data_steinbock_path: Path):
        io.read_neighbors(
            imc_test_data_steinbock_path / "neighbors" / "20210305_NE_mockData1_1.csv"
        )  # TODO

    def test_write_neighbors(self, imc_test_data_steinbock_path: Path):
        pass  # TODO



import shutil
from pathlib import Path
from typing import Generator

import pytest
import requests

_imc_test_data_steinbock_url = (
    "https://github.com/BodenmillerGroup/TestData"
    "/releases/download/v1.0.7/210308_ImcTestData_steinbock.tar.gz"
)
_imc_test_data_steinbock_dir = "datasets/210308_ImcTestData/steinbock"


def _download_and_extract_asset(tmp_dir_path: Path, asset_url: str):
    asset_file_path = tmp_dir_path / "asset.tar.gz"
    response = requests.get(asset_url, stream=True)
    if response.status_code == 200:
        with asset_file_path.open(mode="wb") as f:
            f.write(response.raw.read())
    shutil.unpack_archive(asset_file_path, tmp_dir_path)


@pytest.fixture(scope="session")
def imc_test_data_steinbock_path(
    tmp_path_factory,
) -> Generator[Path, None, None]:
    tmp_dir_path = tmp_path_factory.mktemp("raw")
    _download_and_extract_asset(tmp_dir_path, _imc_test_data_steinbock_url)
    yield tmp_dir_path / Path(_imc_test_data_steinbock_dir)
    shutil.rmtree(tmp_dir_path)



from os import PathLike
from pathlib import Path
from typing import List, Sequence, Union

import click
import click_log
from skimage import measure

from ..._cli.utils import OrderedClickGroup, catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import mosaics


def _collect_tiff_files(
    img_files_or_dirs: Sequence[Union[str, PathLike]]
) -> List[Path]:
    img_files = []
    for img_file_or_dir in img_files_or_dirs:
        if Path(img_file_or_dir).is_file():
            img_files.append(Path(img_file_or_dir))
        else:
            img_files += sorted(Path(img_file_or_dir).rglob("*.tiff"))
    return img_files


@click.group(name="mosaics", cls=OrderedClickGroup, help="Mosaic tiling/stitching")
def mosaics_cmd_group():
    pass


@mosaics_cmd_group.command(name="tile", help="Extract tiles from images")
@click.argument("images", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--size",
    "tile_size",
    type=click.INT,
    required=True,
    help="Tile size (in pixels)",
)
@click.option(
    "--mmap/--no-mmap",
    "mmap",
    default=False,
    show_default=True,
    help="Use memory mapping for reading images",
)
@click.option(
    "-o",
    "tile_dir",
    type=click.Path(file_okay=False),
    required=True,
    help="Path to the tile output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def tile_cmd(images, tile_size, mmap, tile_dir):
    img_files = _collect_tiff_files(images)
    Path(tile_dir).mkdir(exist_ok=True)
    for tile_file, tile in mosaics.try_extract_tiles_from_disk_to_disk(
        img_files, tile_dir, tile_size, mmap=mmap
    ):
        logger.info(tile_file)
        del tile


@mosaics_cmd_group.command(name="stitch", help="Combine tiles into images")
@click.argument("tiles", nargs=-1, type=click.Path(exists=True))
@click.option(
    "--relabel/--no-relabel",
    "relabel",
    default=False,
    show_default=True,
    help="Relabel objects",
)
@click.option(
    "--mmap/--no-mmap",
    "mmap",
    default=False,
    show_default=True,
    help="Use memory mapping for writing images",
)
@click.option(
    "-o",
    "img_dir",
    type=click.Path(file_okay=False),
    required=True,
    help="Path to the tile output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def stitch_cmd(tiles, relabel, mmap, img_dir):
    tile_files = _collect_tiff_files(tiles)
    Path(img_dir).mkdir(exist_ok=True)
    for img_file, img in mosaics.try_stitch_tiles_from_disk_to_disk(
        tile_files, img_dir, mmap=mmap
    ):
        if relabel:
            if img.ndim != 2:
                logger.warning(f"Failed to relabel image with shape {img.shape}")
                continue
            img[:] = measure.label(img)
            img.flush()
        logger.info(img_file)
        del img



import shutil
from pathlib import Path

import numpy as np
import pytest
from steinbock import io
from steinbock.classification import ilastik

ilastik_binary = "/opt/ilastik/run_ilastik.sh"


class TestIlastikClassification:
    def test_list_ilastik_image_files(self, imc_test_data_steinbock_path: Path):
        ilastik.list_ilastik_image_files(
            imc_test_data_steinbock_path / "ilastik_img"
        )  # TODO

    def test_list_ilastik_crop_files(self, imc_test_data_steinbock_path: Path):
        ilastik.list_ilastik_crop_files(
            imc_test_data_steinbock_path / "ilastik_crops"
        )  # TODO

    def test_read_ilastik_image(self, imc_test_data_steinbock_path: Path):
        ilastik.read_ilastik_image(
            imc_test_data_steinbock_path / "ilastik_img" / "20210305_NE_mockData1_1.h5"
        )  # TODO

    def test_read_ilastik_crop(self, imc_test_data_steinbock_path: Path):
        ilastik.read_ilastik_crop(
            imc_test_data_steinbock_path
            / "ilastik_crops"
            / "20210305_NE_mockData1_s0_a1_ac_ilastik_x0_y0_w120_h120.h5"
        )  # TODO

    def test_write_ilastik_image(self, tmp_path: Path):
        ilastik_img = np.array(
            [
                [
                    [1, 2, 3],
                    [4, 5, 6],
                    [7, 8, 9],
                ]
            ],
            dtype=io.img_dtype,
        )
        ilastik.write_ilastik_image(ilastik_img, tmp_path / "ilastik_img.h5")  # TODO

    def test_write_ilastik_crop(self, tmp_path: Path):
        ilastik_crop = np.array(
            [
                [
                    [1, 2, 3],
                    [4, 5, 6],
                    [7, 8, 9],
                ]
            ],
            dtype=io.img_dtype,
        )
        ilastik.write_ilastik_crop(ilastik_crop, tmp_path / "ilastik_crop.h5")  # TODO

    def test_create_ilastik_image(self):
        img = np.array(
            [
                [
                    [1, 2, 3],
                    [4, 5, 6],
                    [7, 8, 9],
                ]
            ],
            dtype=io.img_dtype,
        )
        ilastik_img = ilastik.create_ilastik_image(img)
        expected_ilastik_img = np.array(
            [
                [
                    [100, 200, 300],
                    [400, 500, 600],
                    [700, 800, 900],
                ],
                [
                    [1, 2, 3],
                    [4, 5, 6],
                    [7, 8, 9],
                ],
            ],
            dtype=io.img_dtype,
        )
        assert np.all(ilastik_img == expected_ilastik_img)

    def test_try_create_ilastik_images_from_disk(
        self, imc_test_data_steinbock_path: Path
    ):
        img_files = io.list_image_files(imc_test_data_steinbock_path / "img")
        gen = ilastik.try_create_ilastik_images_from_disk(img_files)
        for img_file, ilastik_img in gen:
            pass  # TODO

    def test_create_ilastik_crop(self):
        ilastik_img = np.array(
            [
                [
                    [100, 200, 300],
                    [400, 500, 600],
                    [700, 800, 900],
                ],
                [
                    [1, 2, 3],
                    [4, 5, 6],
                    [7, 8, 9],
                ],
            ],
            dtype=io.img_dtype,
        )
        rng = np.random.default_rng()
        x, y, ilastik_crop = ilastik.create_ilastik_crop(ilastik_img, 3, rng)
        expected_ilastik_crop = np.array(
            [
                [
                    [100, 200, 300],
                    [400, 500, 600],
                    [700, 800, 900],
                ],
                [
                    [1, 2, 3],
                    [4, 5, 6],
                    [7, 8, 9],
                ],
            ],
            dtype=io.img_dtype,
        )
        assert np.all(ilastik_crop == expected_ilastik_crop)

    def test_try_create_ilastik_crops_from_disk(
        self, imc_test_data_steinbock_path: Path
    ):
        ilastik_img_files = ilastik.list_ilastik_image_files(
            imc_test_data_steinbock_path / "ilastik_img"
        )
        rng = np.random.default_rng(seed=123)
        gen = ilastik.try_create_ilastik_crops_from_disk(ilastik_img_files, 50, rng)
        for ilastik_img_file, x, y, ilastik_crop in gen:
            pass  # TODO

    def test_create_and_save_ilastik_project(
        self, imc_test_data_steinbock_path: Path, tmp_path: Path
    ):
        shutil.copytree(
            imc_test_data_steinbock_path / "ilastik_crops",
            tmp_path / "ilastik_crops",
        )
        ilastik_crop_files = ilastik.list_ilastik_crop_files(tmp_path / "ilastik_crops")
        ilastik.create_and_save_ilastik_project(
            ilastik_crop_files, tmp_path / "pixel_classifier.ilp"
        )  # TODO

    @pytest.mark.skip(reason="Test would take too long")
    @pytest.mark.skipif(
        shutil.which(ilastik_binary) is None, reason="Ilastik is not available"
    )
    def test_run_pixel_classification(
        self, imc_test_data_steinbock_path: Path, tmp_path: Path
    ):
        shutil.copytree(
            imc_test_data_steinbock_path / "ilastik_img",
            tmp_path / "ilastik_img",
        )
        ilastik_img_files = ilastik.list_ilastik_image_files(tmp_path / "ilastik_img")
        ilastik.run_pixel_classification(
            ilastik_binary,
            imc_test_data_steinbock_path / "pixel_classifier.ilp",
            ilastik_img_files,
            tmp_path / "ilastik_probabilities",
        )  # TODO

    def test_try_fix_ilastik_crops_from_disk(self, imc_test_data_steinbock_path: Path):
        pass  # TODO

    def test_fix_ilastik_project_file_inplace(self, imc_test_data_steinbock_path: Path):
        pass  # TODO



from pathlib import Path

import numpy as np
from steinbock import io
from steinbock.measurement import neighbors
from steinbock.measurement.neighbors import NeighborhoodType


class TestNeighborsMeasurement:
    def test_measure_neighbors_centroid(self):
        mask = np.array(
            [
                [1, 1, 0],
                [3, 0, 4],
                [0, 2, 2],
            ],
            dtype=io.mask_dtype,
        )
        neighbors.measure_neighbors(
            mask,
            NeighborhoodType.CENTROID_DISTANCE,
            metric="euclidean",
            dmax=1.0,
        )  # TODO

    def test_measure_neighbors_euclidean_border(self):
        mask = np.array(
            [
                [1, 1, 0],
                [3, 0, 4],
                [0, 2, 2],
            ],
            dtype=io.mask_dtype,
        )
        neighbors.measure_neighbors(
            mask,
            NeighborhoodType.EUCLIDEAN_BORDER_DISTANCE,
            metric="euclidean",
            dmax=1.0,
        )  # TODO

    def test_measure_neighbors_euclidean_pixel_expansion(self):
        mask = np.array(
            [
                [1, 1, 0],
                [3, 0, 4],
                [0, 2, 2],
            ],
            dtype=io.mask_dtype,
        )
        neighbors.measure_neighbors(
            mask,
            NeighborhoodType.EUCLIDEAN_PIXEL_EXPANSION,
            metric="euclidean",
            dmax=1.0,
        )  # TODO

    def test_try_measure_neighbors_from_disk(self, imc_test_data_steinbock_path: Path):
        mask_files = io.list_mask_files(imc_test_data_steinbock_path / "masks")
        gen = neighbors.try_measure_neighbors_from_disk(
            mask_files,
            NeighborhoodType.CENTROID_DISTANCE,
            metric="euclidean",
            dmax=4.0,
        )
        for mask_file, df in gen:
            pass  # TODO



import logging
import re
from os import PathLike
from pathlib import Path
from typing import Generator, NamedTuple, Sequence, Tuple, Union

import numpy as np

from .. import io
from ._utils import SteinbockUtilsException

logger = logging.getLogger(__name__)


class SteinbockMosaicsUtilsException(SteinbockUtilsException):
    pass


def try_extract_tiles_from_disk_to_disk(
    img_files: Sequence[Union[str, PathLike]],
    tile_dir: Union[str, PathLike],
    tile_size: int,
    mmap: bool = False,
) -> Generator[Tuple[Path, np.ndarray], None, None]:
    for img_file in img_files:
        try:
            if mmap:
                img = io.mmap_image(img_file)
            else:
                img = io.read_image(img_file, native_dtype=True)
            if img.shape[-1] % tile_size == 1 or img.shape[-2] % tile_size == 1:
                logger.warning(
                    "Chosen tile size yields UNSTITCHABLE tiles of 1 pixel "
                    f"width or height for image {img_file}"
                )
            for tile_x in range(0, img.shape[-1], tile_size):
                for tile_y in range(0, img.shape[-2], tile_size):
                    tile = img[
                        :,
                        tile_y : (tile_y + tile_size),
                        tile_x : (tile_x + tile_size),
                    ]
                    tile_file = Path(tile_dir) / (
                        f"{Path(img_file).stem}_tx{tile_x}_ty{tile_y}"
                        f"_tw{tile.shape[-1]}_th{tile.shape[-2]}.tiff"
                    )
                    io.write_image(tile, tile_file, ignore_dtype=True)
                    yield tile_file, tile
                    del tile
            del img
        except:
            logger.exception(f"Error extracting tiles: {img_file}")


def try_stitch_tiles_from_disk_to_disk(
    tile_files: Sequence[Union[str, PathLike]],
    img_dir: Union[str, PathLike],
    mmap: bool = False,
) -> Generator[Tuple[str, np.ndarray], None, None]:
    class TileInfo(NamedTuple):
        tile_file: Path
        x: int
        y: int
        width: int
        height: int

    tile_file_stem_pattern = re.compile(
        r"(?P<img_file_stem>.+)_tx(?P<x>\d+)_ty(?P<y>\d+)"
        r"_tw(?P<width>\d+)_th(?P<height>\d+)"
    )
    img_tile_infos = {}
    for tile_file in tile_files:
        m = tile_file_stem_pattern.fullmatch(tile_file.stem)
        if m is None:
            raise SteinbockMosaicsUtilsException(
                f"Malformed tile file name: {tile_file}"
            )
        img_file_stem = m.group("img_file_stem")
        tile_info = TileInfo(
            Path(tile_file),
            int(m.group("x")),
            int(m.group("y")),
            int(m.group("width")),
            int(m.group("height")),
        )
        if img_file_stem not in img_tile_infos:
            img_tile_infos[img_file_stem] = []
        img_tile_infos[img_file_stem].append(tile_info)
    for img_file_stem, tile_infos in img_tile_infos.items():
        img_file = Path(img_dir) / f"{img_file_stem}.tiff"
        try:
            tile = io.read_image(tile_infos[0].tile_file, native_dtype=True)
            img_shape = (
                tile.shape[0],
                max(ti.y + ti.height for ti in tile_infos),
                max(ti.x + ti.width for ti in tile_infos),
            )
            if mmap:
                img = io.mmap_image(
                    img_file, mode="r+", shape=img_shape, dtype=tile.dtype
                )
            else:
                img = np.zeros(img_shape, dtype=tile.dtype)
            for i, tile_info in enumerate(tile_infos):
                if i > 0:
                    tile = io.read_image(tile_info.tile_file, native_dtype=True)
                img[
                    :,
                    tile_info.y : tile_info.y + tile_info.height,
                    tile_info.x : tile_info.x + tile_info.width,
                ] = tile
                if mmap:
                    img.flush()
            if not mmap:
                io.write_image(img, img_file, ignore_dtype=True)
            yield img_file, img
            del img
        except:
            logger.exception(f"Error stitching tiles: {img_file}")



from pathlib import Path

import pandas as pd
from steinbock import io
from steinbock.export import graphs


class TestGraphsExport:
    def test_convert_to_networkx(self):
        neighbors = pd.DataFrame(
            data={
                "Object": [1, 2],
                "Neighbor": [2, 1],
                "Distance": [1.0, 1.0],
            }
        )
        neighbors["Object"] = neighbors["Object"].astype(io.mask_dtype)
        neighbors["Neighbor"] = neighbors["Neighbor"].astype(io.mask_dtype)
        intensities = pd.DataFrame(
            data={
                "Channel 1": [1.0, 2.0],
                "Channel 2": [100.0, 200.0],
            },
            index=pd.Index([1, 2], name="Object", dtype=io.mask_dtype),
        )
        graphs.convert_to_networkx(neighbors, intensities)  # TODO

    def test_try_convert_to_networkx_from_disk(
        self, imc_test_data_steinbock_path: Path
    ):
        neighbors_files = io.list_neighbors_files(
            imc_test_data_steinbock_path / "neighbors"
        )
        intensities_files = io.list_data_files(
            imc_test_data_steinbock_path / "intensities",
            base_files=neighbors_files,
        )
        gen = graphs.try_convert_to_networkx_from_disk(
            neighbors_files, intensities_files
        )
        for neighbors_file, (intensities_file,), graph in gen:
            pass  # TODO



import sys
from pathlib import Path

import click
import click_log

from ..._cli.utils import OrderedClickGroup, catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import cellprofiler


@click.group(
    name="cellprofiler",
    cls=OrderedClickGroup,
    help="Segment objects in probability images using CellProfiler",
)
def cellprofiler_cmd_group():
    pass


@cellprofiler_cmd_group.command(
    name="prepare", help="Prepare a CellProfiler segmentation pipeline"
)
@click.option(
    "-o",
    "segmentation_pipeline_file",
    type=click.Path(dir_okay=False),
    default="cell_segmentation.cppipe",
    show_default=True,
    help="Path to the CellProfiler segmentation pipeline output file",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def prepare_cmd(segmentation_pipeline_file):
    cellprofiler.create_and_save_segmentation_pipeline(segmentation_pipeline_file)
    logger.info(segmentation_pipeline_file)


@cellprofiler_cmd_group.command(
    name="run", help="Run a object segmentation batch using CellProfiler"
)
@click.option(
    "--cellprofiler",
    "cellprofiler_binary",
    type=click.STRING,
    default="cellprofiler",
    show_default=True,
    help="CellProfiler binary",
)
@click.option(
    "--plugins-directory",
    "cellprofiler_plugin_dir",
    type=click.Path(file_okay=False),
    default="/opt/cellprofiler_plugins",
    show_default=True,
    help="Path to the CellProfiler plugin directory",
)
@click.option(
    "--pipe",
    "segmentation_pipeline_file",
    type=click.Path(exists=True, dir_okay=False),
    default="cell_segmentation.cppipe",
    show_default=True,
    help="Path to the CellProfiler segmentation pipeline file",
)
@click.option(
    "--probabs",
    "probabilities_dir",
    type=click.Path(exists=True, file_okay=False),
    default="ilastik_probabilities",
    show_default=True,
    help="Path to the probabilities directory",
)
@click.option(
    "-o",
    "mask_dir",
    type=click.Path(file_okay=False),
    default="masks",
    show_default=True,
    help="Path to the mask output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def run_cmd(
    cellprofiler_binary,
    cellprofiler_plugin_dir,
    segmentation_pipeline_file,
    probabilities_dir,
    mask_dir,
):
    if probabilities_dir not in ("ilastik_probabilities"):
        logger.warning(
            "When using custom probabilities from unknown origins, "
            "make sure to adapt the CellProfiler pipeline accordingly"
        )
    Path(mask_dir).mkdir(exist_ok=True)
    result = cellprofiler.try_segment_objects(
        cellprofiler_binary,
        segmentation_pipeline_file,
        probabilities_dir,
        mask_dir,
        cellprofiler_plugin_dir=cellprofiler_plugin_dir,
    )
    sys.exit(result.returncode)



import os
import logging
import subprocess
import sys
from functools import wraps
from pathlib import Path


logger = logging.getLogger(__name__.rpartition(".")[0])


def run_captured(
    args, *popen_args, file=sys.stdout, **popen_kwargs
) -> subprocess.CompletedProcess:
    with subprocess.Popen(
        args,
        *popen_args,
        bufsize=0,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        **popen_kwargs,
    ) as process:
        for c in iter(lambda: process.stdout.read(1), b""):
            file.buffer.write(c)
        process.wait()
    return subprocess.CompletedProcess(
        process.args, process.returncode, process.stdout, process.stderr
    )


def check_x11(func):
    @wraps(func)
    def check_x11_wrapper(*args, **kwargs):
        if "DISPLAY" not in os.environ:
            logger.warning("X11 required; did you set $DISPLAY?")
        x11_path = Path("/tmp/.X11-unix")
        if not x11_path.exists():
            logger.warning(f"X11 required; did you bind-mount {x11_path}?")
        xauth_path = Path("~/.Xauthority").expanduser()
        if not xauth_path.exists():
            logger.warning(f"X11 required; did you bind-mount {xauth_path}?")
        return func(*args, **kwargs)

    return check_x11_wrapper


def use_ilastik_env(func):
    @wraps(func)
    def use_ilastik_env_wrapper(*args, **kwargs):
        if "ilastik_env" not in kwargs:
            kwargs["ilastik_env"] = os.environ.copy()
        kwargs["ilastik_env"].pop("PYTHONPATH", None)
        kwargs["ilastik_env"].pop("PYTHONHOME", None)
        kwargs["ilastik_env"].pop("LD_LIBRARY_PATH", None)
        return func(*args, **kwargs)

    return use_ilastik_env_wrapper



import sys
from pathlib import Path

import click
import click_log
import numpy as np
import tifffile

from ... import io
from ..._cli.utils import OrderedClickGroup, catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import cellprofiler


@click.group(
    name="cellprofiler",
    cls=OrderedClickGroup,
    help="Run a CellProfiler measurement pipeline (legacy)",
)
def cellprofiler_cmd_group():
    pass


@cellprofiler_cmd_group.command(
    name="prepare", help="Prepare a CellProfiler measurement pipeline"
)
@click.option(
    "--img",
    "img_dir",
    type=click.Path(exists=True, file_okay=False),
    default="img",
    show_default=True,
    help="Path to the image directory",
)
@click.option(
    "--masks",
    "mask_dir",
    type=click.Path(exists=True, file_okay=False),
    default="masks",
    show_default=True,
    help="Path to the mask directory",
)
@click.option(
    "--panel",
    "panel_file",
    type=click.Path(exists=True, dir_okay=False),
    default="panel.csv",
    show_default=True,
    help="Path to the panel file",
)
@click.option(
    "--pipeout",
    "measurement_pipeline_file",
    type=click.Path(dir_okay=False),
    default="cell_measurement.cppipe",
    show_default=True,
    help="Path to the CellProfiler measurement pipeline output file",
)
@click.option(
    "--dataout",
    "cpdata_dir",
    type=click.Path(file_okay=False),
    default="cellprofiler_input",
    show_default=True,
    help="Path to the CellProfiler input directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def prepare_cmd(
    img_dir,
    mask_dir,
    panel_file,
    measurement_pipeline_file,
    cpdata_dir,
):
    panel = io.read_panel(panel_file)
    img_files = io.list_image_files(img_dir)
    mask_files = io.list_mask_files(mask_dir, base_files=img_files)
    Path(cpdata_dir).mkdir(exist_ok=True)
    for img_file, mask_file in zip(img_files, mask_files):
        img = io.read_image(img_file, native_dtype=True)
        cp_img_file = Path(cpdata_dir) / img_file.name
        tifffile.imwrite(
            cp_img_file,
            data=io._to_dtype(img, np.uint16)[
                np.newaxis, np.newaxis, :, :, :, np.newaxis
            ],
            imagej=True,
        )
        logger.info(cp_img_file)
        del img
        mask = io.read_mask(mask_file, native_dtype=True)
        cp_mask_file = Path(cpdata_dir) / f"{mask_file.stem}_mask{mask_file.suffix}"
        tifffile.imwrite(
            cp_mask_file,
            data=io._to_dtype(mask, np.uint16)[
                np.newaxis, np.newaxis, np.newaxis, :, :, np.newaxis
            ],
            imagej=True,
        )
        logger.info(cp_mask_file)
        del mask
    cellprofiler.create_and_save_measurement_pipeline(
        measurement_pipeline_file, len(panel.index)
    )
    logger.info(measurement_pipeline_file)


@cellprofiler_cmd_group.command(
    name="run", help="Run an object measurement batch using CellProfiler"
)
@click.option(
    "--cellprofiler",
    "cellprofiler_binary",
    type=click.STRING,
    default="cellprofiler",
    show_default=True,
    help="CellProfiler binary",
)
@click.option(
    "--plugins-directory",
    "cellprofiler_plugin_dir",
    type=click.Path(file_okay=False),
    default="/opt/cellprofiler_plugins",
    show_default=True,
    help="Path to the CellProfiler plugin directory",
)
@click.option(
    "--pipe",
    "measurement_pipeline_file",
    type=click.Path(exists=True, dir_okay=False),
    default="cell_measurement.cppipe",
    show_default=True,
    help="Path to the CellProfiler measurement pipeline file",
)
@click.option(
    "--data",
    "cpdata_dir",
    type=click.Path(exists=True, file_okay=False),
    default="cellprofiler_input",
    show_default=True,
    help="Path to the CellProfiler input directory",
)
@click.option(
    "-o",
    "cpout_dir",
    type=click.Path(file_okay=False),
    default="cellprofiler_output",
    show_default=True,
    help="Path to the CellProfiler output directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def run_cmd(
    cellprofiler_binary,
    cellprofiler_plugin_dir,
    measurement_pipeline_file,
    cpdata_dir,
    cpout_dir,
):
    Path(cpout_dir).mkdir(exist_ok=True)
    result = cellprofiler.try_measure_objects(
        cellprofiler_binary,
        measurement_pipeline_file,
        cpdata_dir,
        cpout_dir,
        cellprofiler_plugin_dir=cellprofiler_plugin_dir,
    )
    sys.exit(result.returncode)



import logging
from enum import Enum
from functools import partial
from os import PathLike
from pathlib import Path
from typing import Generator, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from scipy.ndimage import measurements

from .. import io

logger = logging.getLogger(__name__)


class IntensityAggregation(Enum):
    """"""

    SUM = partial(measurements.sum_labels)
    MIN = partial(measurements.minimum)
    MAX = partial(measurements.maximum)
    MEAN = partial(measurements.mean)
    MEDIAN = partial(measurements.median)
    STD = partial(measurements.standard_deviation)
    VAR = partial(measurements.variance)


def measure_intensites(
    img: np.ndarray,
    mask: np.ndarray,
    channel_names: Sequence[str],
    intensity_aggregation: IntensityAggregation,
) -> pd.DataFrame:
    object_ids = np.unique(mask[mask != 0])
    data = {
        channel_name: intensity_aggregation.value(img[i], labels=mask, index=object_ids)
        for i, channel_name in enumerate(channel_names)
    }
    return pd.DataFrame(
        data=data,
        index=pd.Index(object_ids, dtype=io.mask_dtype, name="Object"),
    )


def try_measure_intensities_from_disk(
    img_files: Sequence[Union[str, PathLike]],
    mask_files: Sequence[Union[str, PathLike]],
    channel_names: Sequence[str],
    intensity_aggregation: IntensityAggregation,
    mmap: bool = False,
) -> Generator[Tuple[Path, Path, pd.DataFrame], None, None]:
    for img_file, mask_file in zip(img_files, mask_files):
        try:
            if mmap:
                img = io.mmap_image(img_file)
                mask = io.mmap_mask(mask_file)
            else:
                img = io.read_image(img_file)
                mask = io.read_mask(mask_file)
            intensities = measure_intensites(
                img, mask, channel_names, intensity_aggregation
            )
            del img, mask
            yield Path(img_file), Path(mask_file), intensities
            del intensities
        except:
            logger.exception(f"Error measuring intensities in {img_file}")



import logging
import os
import re
from os import PathLike
from pathlib import Path
from typing import List, Optional, Sequence, Union

import numpy as np
import pandas as pd
import tifffile

from ._steinbock import SteinbockException

logger = logging.getLogger(__name__)
img_dtype = np.dtype(os.environ.get("STEINBOCK_IMG_DTYPE", "float32"))
mask_dtype = np.dtype(os.environ.get("STEINBOCK_MASK_DTYPE", "uint16"))


class SteinbockIOException(SteinbockException):
    pass


def _as_path_with_suffix(path: Union[str, PathLike], suffix: str) -> Path:
    path = Path(path)
    if re.fullmatch(r".+\.ome\.[^.]+", path.name, flags=re.IGNORECASE):
        stem, ome_suffix, suffix = path.name.rpartition(".ome.")
        if ome_suffix:
            path = path.with_name(f"{stem}.{suffix}")
        stem, ome_suffix, suffix = path.name.rpartition(".OME.")
        if ome_suffix:
            path = path.with_name(f"{stem}.{suffix}")
    return path.with_suffix(suffix)


def _to_dtype(src: np.ndarray, dst_dtype: np.dtype) -> np.ndarray:
    if src.dtype == dst_dtype:
        return src
    src_is_int = np.issubdtype(src.dtype, np.integer)
    dst_is_int = np.issubdtype(dst_dtype, np.integer)
    if not src_is_int and dst_is_int:
        src = np.around(src)
    if src_is_int:
        src_info = np.iinfo(src.dtype)
    else:
        src_info = np.finfo(src.dtype)
    if dst_is_int:
        dst_info = np.iinfo(dst_dtype)
    else:
        dst_info = np.finfo(dst_dtype)
    if src_info.min < dst_info.min or src_info.max > dst_info.max:
        src = np.clip(src, dst_info.min, dst_info.max)
    return src.astype(dst_dtype)


def _list_related_files(
    base_files: Sequence[Union[str, PathLike]],
    related_dir: Union[str, PathLike],
    related_suffix: str,
) -> List[Path]:
    related_files = []
    for base_file in base_files:
        related_file = _as_path_with_suffix(
            Path(related_dir) / Path(base_file).name, related_suffix
        )
        if not related_file.exists():
            raise SteinbockIOException() from FileNotFoundError(related_file)
        related_files.append(related_file)
    return related_files


def read_panel(
    panel_file: Union[str, PathLike], unfiltered: bool = False
) -> pd.DataFrame:
    panel = pd.read_csv(
        panel_file,
        sep=",|;",
        dtype={
            "channel": pd.StringDtype(),
            "name": pd.StringDtype(),
            "keep": pd.BooleanDtype(),
        },
        engine="python",
        true_values=["1"],
        false_values=["0"],
    )
    for required_col in ("channel", "name"):
        if required_col not in panel:
            raise SteinbockIOException(
                f"Missing '{required_col}' column in {panel_file}"
            )
    for notnan_col in ("channel", "keep"):
        if notnan_col in panel and panel[notnan_col].isna().any():
            raise SteinbockIOException(
                f"Missing values for '{notnan_col}' in {panel_file}"
            )
    for unique_col in ("channel", "name"):
        if unique_col in panel:
            if panel[unique_col].dropna().duplicated().any():
                raise SteinbockIOException(
                    f"Duplicated values for '{unique_col}' in {panel_file}"
                )
    if not unfiltered and "keep" in panel:
        panel = panel.loc[panel["keep"].astype(bool), :]
    return panel


def write_panel(panel: pd.DataFrame, panel_file: Union[str, PathLike]) -> None:
    panel = panel.copy()
    for col in panel.columns:
        if panel[col].convert_dtypes().dtype == pd.BooleanDtype():
            panel[col] = panel[col].astype(pd.UInt8Dtype())
    panel.to_csv(panel_file, index=False)


def list_image_files(
    img_dir: Union[str, PathLike],
    base_files: Optional[Sequence[Union[str, PathLike]]] = None,
) -> List[Path]:
    if base_files is not None:
        return _list_related_files(base_files, img_dir, ".tiff")
    return sorted(Path(img_dir).rglob("*.tiff"))


def _fix_image_shape(img_file: Union[str, PathLike], img: np.ndarray) -> np.ndarray:
    if img.ndim == 2:
        img = img[np.newaxis, :, :]
    elif img.ndim == 5:
        size_t, size_z, size_c, size_y, size_x = img.shape
        if size_t != 1 or size_z != 1:
            raise SteinbockIOException(
                f"{img_file}: unsupported TZCYX shape {img.shape}"
            )
        img = img[0, 0, :, :, :]
    elif img.ndim == 6:
        size_t, size_z, size_c, size_y, size_x, size_s = img.shape
        if size_t != 1 or size_z != 1 or size_s != 1:
            raise SteinbockIOException(
                f"{img_file}: unsupported TZCYXS shape {img.shape}"
            )
        img = img[0, 0, :, :, :, 0]
    elif img.ndim != 3:
        raise SteinbockIOException(
            f"{img_file}: unsupported number of dimensions ({img.ndim})"
        )
    return img


def read_image(
    img_file: Union[str, PathLike],
    native_dtype: bool = False,
) -> np.ndarray:
    img = tifffile.imread(img_file, squeeze=False)
    img = _fix_image_shape(img_file, img)
    if not native_dtype:
        img = _to_dtype(img, img_dtype)
    return img


def mmap_image(img_file: Union[str, PathLike], mode="r", **kwargs) -> np.ndarray:
    if "imagej" not in kwargs and mode == "r+":
        kwargs["imagej"] = True
    img_exists = Path(img_file).exists()
    img = tifffile.memmap(img_file, mode=mode, **kwargs)
    if img_exists:
        if img.dtype != img_dtype:
            logger.warning(
                "Data type of memory-mapped image file %s (%s) is not %s",
                img_file,
                img.dtype,
                img_dtype,
            )
        img = _fix_image_shape(img_file, img)
    return img


def write_image(
    img: np.ndarray,
    img_file: Union[str, PathLike],
    ignore_dtype: bool = False,
) -> None:
    if not ignore_dtype:
        img = _to_dtype(img, img_dtype)
    tifffile.imwrite(
        img_file,
        data=img[np.newaxis, np.newaxis, :, :, :, np.newaxis],
        imagej=True,
    )


def read_image_info(image_info_file: Union[str, PathLike]) -> pd.DataFrame:
    image_info = pd.read_csv(
        image_info_file,
        sep=",|;",
        dtype={
            "image": pd.StringDtype(),
            "width_px": pd.UInt16Dtype(),
            "height_px": pd.UInt16Dtype(),
            "num_channels": pd.UInt8Dtype(),
        },
        engine="python",
    )
    for required_col in ("image", "width_px", "height_px", "num_channels"):
        if required_col not in image_info:
            raise SteinbockIOException(
                f"Missing '{required_col}' column in {image_info_file}"
            )
    for notnan_col in ("image", "width_px", "height_px", "num_channels"):
        if notnan_col in image_info and image_info[notnan_col].isna().any():
            raise SteinbockIOException(
                f"Missing values for '{notnan_col}' in {image_info_file}"
            )
    for unique_col in ("image",):
        if unique_col in image_info:
            if image_info[unique_col].dropna().duplicated().any():
                raise SteinbockIOException(
                    f"Duplicated values for '{unique_col}'" f" in {image_info_file}"
                )
    return image_info


def write_image_info(
    image_info: pd.DataFrame, image_info_file: Union[str, PathLike]
) -> None:
    image_info.to_csv(image_info_file, index=False)


def list_mask_files(
    mask_dir: Union[str, PathLike],
    base_files: Optional[Sequence[Union[str, PathLike]]] = None,
) -> List[Path]:
    if base_files is not None:
        return _list_related_files(base_files, mask_dir, ".tiff")
    return sorted(Path(mask_dir).rglob("*.tiff"))


def _fix_mask_shape(mask_file: Union[str, PathLike], mask: np.ndarray) -> np.ndarray:
    if mask.ndim == 5:
        size_t, size_z, size_c, size_y, size_x = mask.shape
        if size_t != 1 or size_z != 1 or size_c != 1:
            raise SteinbockIOException(
                f"{mask_file}: unsupported TZCYX shape {mask.shape}"
            )
        mask = mask[0, 0, 0, :, :]
    elif mask.ndim == 6:
        size_t, size_z, size_c, size_y, size_x, size_s = mask.shape
        if size_t != 1 or size_z != 1 or size_c != 1 or size_s != 1:
            raise SteinbockIOException(
                f"{mask_file}: unsupported TZCYXS shape {mask.shape}"
            )
        mask = mask[0, 0, 0, :, :, 0]
    elif mask.ndim != 2:
        raise SteinbockIOException(
            f"{mask_file}: unsupported number of dimensions ({mask.ndim})"
        )
    return mask


def read_mask(
    mask_file: Union[str, PathLike],
    native_dtype: bool = False,
) -> np.ndarray:
    mask = tifffile.imread(mask_file, squeeze=False)
    mask = _fix_mask_shape(mask_file, mask)
    if not native_dtype:
        mask = _to_dtype(mask, mask_dtype)
    return mask


def mmap_mask(mask_file: Union[str, PathLike], mode="r", **kwargs) -> np.ndarray:
    if "imagej" not in kwargs and mode == "r+":
        kwargs["imagej"] = True
    mask_exists = Path(mask_file).exists()
    mask = tifffile.memmap(mask_file, mode=mode, **kwargs)
    if mask_exists:
        if mask.dtype != mask_dtype:
            logger.warning(
                "Data type of memory-mapped mask file %s (%s) is not %s",
                mask_file,
                mask.dtype,
                mask_dtype,
            )
        mask = _fix_mask_shape(mask_file, mask)
    return mask


def write_mask(
    mask: np.ndarray,
    mask_file: Union[str, PathLike],
    ignore_dtype: bool = False,
) -> None:
    if not ignore_dtype:
        mask = _to_dtype(mask, mask_dtype)
    tifffile.imwrite(
        mask_file,
        data=mask[np.newaxis, np.newaxis, np.newaxis, :, :, np.newaxis],
        imagej=True,
    )


def list_data_files(
    data_dir: Union[str, PathLike],
    base_files: Optional[Sequence[Union[str, PathLike]]] = None,
) -> List[Path]:
    if base_files is not None:
        return _list_related_files(base_files, data_dir, ".csv")
    return sorted(Path(data_dir).rglob("*.csv"))


def read_data(data_file: Union[str, PathLike]) -> pd.DataFrame:
    return pd.read_csv(data_file, sep=",|;", index_col="Object", engine="python")


def write_data(data: pd.DataFrame, data_file: Union[str, PathLike]) -> None:
    data = data.reset_index()
    data.to_csv(data_file, index=False)


def list_neighbors_files(
    neighbors_dir: Union[str, PathLike],
    base_files: Optional[Sequence[Union[str, PathLike]]] = None,
) -> List[Path]:
    if base_files is not None:
        return _list_related_files(base_files, neighbors_dir, ".csv")
    return sorted(Path(neighbors_dir).rglob("*.csv"))


def read_neighbors(neighbors_file: Union[str, PathLike]) -> pd.DataFrame:
    return pd.read_csv(
        neighbors_file,
        sep=",|;",
        usecols=["Object", "Neighbor", "Distance"],
        dtype={
            "Object": mask_dtype,
            "Neighbor": mask_dtype,
            "Distance": np.float32,
        },
        engine="python",
    )


def write_neighbors(
    neighbors: pd.DataFrame, neighbors_file: Union[str, PathLike]
) -> None:
    neighbors = neighbors.loc[:, ["Object", "Neighbor", "Distance"]].astype(
        {
            "Object": mask_dtype,
            "Neighbor": mask_dtype,
            "Distance": np.float32,
        }
    )
    neighbors.to_csv(neighbors_file, index=False)



import logging
from os import PathLike
from pathlib import Path
from typing import Generator, List, Sequence, Tuple, Union

import imageio
import numpy as np
import pandas as pd

from .. import io
from ._preprocessing import SteinbockPreprocessingException

logger = logging.getLogger(__name__)


class SteinbockExternalPreprocessingException(SteinbockPreprocessingException):
    pass


def _read_external_image(ext_img_file: Union[str, PathLike]) -> np.ndarray:
    # try tifffile-based reading first, for memory reasons
    try:
        return io.read_image(ext_img_file, native_dtype=True)
    except:
        pass  # skipped intentionally
    ext_img = imageio.volread(ext_img_file)
    orig_img_shape = ext_img.shape
    while ext_img.ndim > 3 and ext_img.shape[0] == 1:
        ext_img = np.squeeze(ext_img, axis=0)
    while ext_img.ndim > 3 and ext_img.shape[-1] == 1:
        ext_img = np.squeeze(ext_img, axis=-1)
    if ext_img.ndim == 2:
        ext_img = ext_img[np.newaxis, :, :]
    elif ext_img.ndim != 3:
        raise SteinbockExternalPreprocessingException(
            f"Unsupported shape {orig_img_shape} for image {ext_img_file}"
        )
    return ext_img


def list_image_files(ext_img_dir: Union[str, PathLike]) -> List[Path]:
    return sorted(Path(ext_img_dir).rglob("*.*"))


def create_panel_from_image_files(
    ext_img_files: Sequence[Union[str, PathLike]]
) -> pd.DataFrame:
    num_channels = None
    for ext_img_file in ext_img_files:
        try:
            ext_img = _read_external_image(ext_img_file)
            num_channels = ext_img.shape[0]
            break
        except:
            pass  # skipped intentionally
    if num_channels is None:
        raise SteinbockExternalPreprocessingException("No valid images found")
    panel = pd.DataFrame(
        data={
            "channel": range(1, num_channels + 1),
            "name": np.nan,
            "keep": True,
            "ilastik": range(1, num_channels + 1),
            "deepcell": np.nan,
        },
    )
    panel["channel"] = panel["channel"].astype(pd.StringDtype())
    panel["name"] = panel["name"].astype(pd.StringDtype())
    panel["keep"] = panel["keep"].astype(pd.BooleanDtype())
    panel["ilastik"] = panel["ilastik"].astype(pd.UInt8Dtype())
    panel["deepcell"] = panel["deepcell"].astype(pd.UInt8Dtype())
    return panel


def try_preprocess_images_from_disk(
    ext_img_files: Sequence[Union[str, PathLike]]
) -> Generator[Tuple[Path, np.ndarray], None, None]:
    for ext_img_file in ext_img_files:
        try:
            img = _read_external_image(ext_img_file)
        except:
            logger.warning(f"Unsupported file format: {ext_img_file}")
            continue
        yield ext_img_file, img
        del img



import logging
from os import PathLike
from pathlib import Path
from typing import Generator, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from skimage.measure import regionprops_table

from .. import io

logger = logging.getLogger(__name__)


def measure_regionprops(
    img: np.ndarray, mask: np.ndarray, skimage_regionprops: Sequence[str]
) -> pd.DataFrame:
    skimage_regionprops = list(skimage_regionprops)
    if "label" not in skimage_regionprops:
        skimage_regionprops.insert(0, "label")
    data = regionprops_table(
        mask,
        intensity_image=np.moveaxis(img, 0, -1),
        properties=skimage_regionprops,
    )
    object_ids = data.pop("label")
    return pd.DataFrame(
        data=data,
        index=pd.Index(object_ids, dtype=io.mask_dtype, name="Object"),
    )


def try_measure_regionprops_from_disk(
    img_files: Sequence[Union[str, PathLike]],
    mask_files: Sequence[Union[str, PathLike]],
    skimage_regionprops: Sequence[str],
    mmap: bool = False,
) -> Generator[Tuple[Path, Path, pd.DataFrame], None, None]:
    for img_file, mask_file in zip(img_files, mask_files):
        try:
            if mmap:
                img = io.mmap_image(img_file)
                mask = io.mmap_mask(mask_file)
            else:
                img = io.read_image(img_file)
                mask = io.read_mask(mask_file)
            regionprops = measure_regionprops(img, mask, skimage_regionprops)
            del img, mask
            yield Path(img_file), Path(mask_file), regionprops
            del regionprops
        except:
            logger.exception(f"Error measuring regionprops in {img_file}")



from pathlib import Path

import click
import click_log
import pandas as pd
from anndata import concat as anndata_concat
from fcswrite import write_fcs

from ... import io
from ..._cli.utils import SteinbockCLIException, catch_exception, logger
from ..._steinbock import SteinbockException
from ..._steinbock import logger as steinbock_logger
from .. import data


@click.command(name="csv", help="Merge and export object data to CSV")
@click.argument("data_dirs", nargs=-1, type=click.Path(exists=True, file_okay=False))
@click.option(
    "--concat/--no-concat",
    "concatenate",
    default=True,
    show_default=True,
    help="Concatenate all files into a single file",
)
@click.option(
    "-o",
    "csv_file_or_dir",
    type=click.Path(),
    required=True,
    help="Path to the CSV output file or directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def csv_cmd(data_dirs, concatenate, csv_file_or_dir):
    if not data_dirs:
        return  # empty variadic argument, gracefully degrade into noop
    if concatenate and Path(csv_file_or_dir).is_dir():
        raise SteinbockCLIException("Specify a single output file when concatenating")
    elif not concatenate and Path(csv_file_or_dir).is_file():
        raise SteinbockCLIException("Output directory is a file")
    first_data_files = io.list_data_files(data_dirs[0])
    data_file_lists = [first_data_files]
    for data_dir in data_dirs[1:]:
        data_files = io.list_data_files(data_dir, base_files=first_data_files)
        data_file_lists.append(data_files)
    if not concatenate:
        Path(csv_file_or_dir).mkdir(exist_ok=True)
    for i, (img_file_name, data_files, df) in enumerate(
        data.try_convert_to_dataframe_from_disk(*data_file_lists)
    ):
        df.reset_index(inplace=True)
        if concatenate:
            df.insert(0, "Image", img_file_name)
            if i == 0:
                df.to_csv(csv_file_or_dir, index=False)
            else:
                df.to_csv(csv_file_or_dir, header=False, index=False, mode="a")
        else:
            csv_file = io._as_path_with_suffix(
                Path(csv_file_or_dir) / img_file_name, ".csv"
            )
            df.to_csv(csv_file, index=False)
            logger.info(csv_file)
        del df
    if concatenate:
        logger.info(csv_file_or_dir)


@click.command(name="fcs", help="Merge and export object data to FCS")
@click.argument("data_dirs", nargs=-1, type=click.Path(exists=True, file_okay=False))
@click.option(
    "--concat/--no-concat",
    "concatenate",
    default=True,
    show_default=True,
    help="Concatenate all files into a single file",
)
@click.option(
    "-o",
    "fcs_file_or_dir",
    type=click.Path(),
    required=True,
    help="Path to the FCS output file or directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def fcs_cmd(data_dirs, concatenate, fcs_file_or_dir):
    if not data_dirs:
        return  # empty variadic argument, gracefully degrade into noop
    if concatenate and Path(fcs_file_or_dir).is_dir():
        raise SteinbockCLIException("Specify a single output file when concatenating")
    elif not concatenate and Path(fcs_file_or_dir).is_file():
        raise SteinbockCLIException("Output directory is a file")
    first_data_files = io.list_data_files(data_dirs[0])
    data_file_lists = [first_data_files]
    for data_dir in data_dirs[1:]:
        data_files = io.list_data_files(data_dir, base_files=first_data_files)
        data_file_lists.append(data_files)
    if concatenate:
        logger.warning(
            "The fcswrite package currently does not support on-disk concatenation; "
            "all files will be loaded into memory"
        )
    else:
        Path(fcs_file_or_dir).mkdir(exist_ok=True)
    dfs = []
    for (
        img_file_name,
        data_files,
        df,
    ) in data.try_convert_to_dataframe_from_disk(*data_file_lists):
        if concatenate:
            dfs.append(df)
        else:
            fcs_file = io._as_path_with_suffix(
                Path(fcs_file_or_dir) / img_file_name, ".fcs"
            )
            write_fcs(fcs_file, df.columns.values, df.values)
            logger.info(fcs_file)
            del df
    if concatenate:
        df = pd.concat(dfs, ignore_index=True, copy=False)
        write_fcs(fcs_file_or_dir, df.columns.values, df.values)
        logger.info(fcs_file_or_dir)


@click.command(name="anndata", help="Merge and export object data to AnnData")
@click.option(
    "--intensities",
    "intensities_dir",
    required=True,
    type=click.Path(exists=True, file_okay=False),
    help="Path to the intensities directory",
)
@click.option(
    "--data",
    "data_dirs",
    multiple=True,
    type=click.Path(exists=True, file_okay=False),
    help="Path to the object data directory",
)
@click.option(
    "--neighbors",
    "neighbors_dir",
    type=click.Path(exists=True, file_okay=False),
    help="Path to the neighbors directory",
)
# @click.option(
#     "--panel",
#     "panel_file",
#     type=click.Path(dir_okay=False),
#     default="panel.csv",
#     show_default=True,
#     help="Path to the panel file",
# )
# @click.option(
#     "--info",
#     "image_info_file",
#     type=click.Path(dir_okay=False),
#     default="images.csv",
#     show_default=True,
#     help="Path to the image information file",
# )
@click.option(
    "--concat/--no-concat",
    "concatenate",
    default=True,
    show_default=True,
    help="Concatenate all files into a single file",
)
@click.option(
    "--format",
    "anndata_format",
    type=click.Choice(["h5ad", "loom", "zarr"], case_sensitive=False),
    default="h5ad",
    show_default=True,
    help="AnnData file format to use",
)
@click.option(
    "-o",
    "anndata_file_or_dir",
    type=click.Path(),
    required=True,
    help="Path to the AnnData output file or directory",
)
@click_log.simple_verbosity_option(logger=steinbock_logger)
@catch_exception(handle=SteinbockException)
def anndata_cmd(
    intensities_dir,
    data_dirs,
    neighbors_dir,
    # panel_file,
    # image_info_file,
    concatenate,
    anndata_format,
    anndata_file_or_dir,
):
    if anndata_format == "h5ad":

        def write_anndata(adata, anndata_file, ignore_suffix=False):
            suffix = Path(anndata_file).suffix if ignore_suffix else ".h5ad"
            anndata_file = io._as_path_with_suffix(anndata_file, suffix)
            adata.write_h5ad(anndata_file)
            return anndata_file

    elif anndata_format == "loom":

        def write_anndata(adata, anndata_file, ignore_suffix=False):
            suffix = Path(anndata_file).suffix if ignore_suffix else ".loom"
            anndata_file = io._as_path_with_suffix(anndata_file, suffix)
            adata.write_loom(anndata_file)
            return anndata_file

    elif anndata_format == "zarr":

        def write_anndata(adata, anndata_file, ignore_suffix=False):
            suffix = Path(anndata_file).suffix if ignore_suffix else ".zarr"
            anndata_file = io._as_path_with_suffix(anndata_file, suffix)
            adata.write_zarr(anndata_file)
            return anndata_file

    else:
        raise NotImplementedError()

    intensities_files = io.list_data_files(intensities_dir)
    data_file_lists = [
        io.list_data_files(data_dir, base_files=intensities_files)
        for data_dir in data_dirs
        if Path(data_dir).is_dir()
    ]
    neighbors_files = None
    if neighbors_dir is not None and Path(neighbors_dir).is_dir():
        neighbors_files = io.list_neighbors_files(
            neighbors_dir, base_files=intensities_files
        )
    # panel = None
    # if panel_file is not None and Path(panel_file).is_file():
    #     panel = io.read_panel(panel_file)
    # image_info = None
    # if image_info_file is not None and Path(image_info_file).is_file():
    #     image_info = io.read_image_info(image_info_file)
    if concatenate and Path(anndata_file_or_dir).is_dir():
        raise SteinbockCLIException("Specify a single output file when concatenating")
    elif not concatenate and Path(anndata_file_or_dir).is_file():
        raise SteinbockCLIException("Output directory is a file")
    if concatenate:
        logger.warning(
            "The anndata package currently does not support on-disk; "
            "all files will be loaded into memory"
        )
    else:
        Path(anndata_file_or_dir).mkdir(exist_ok=True)
    adatas = {}
    for (
        img_file_name,
        intensities_file,
        data_files,
        neighbors_file,
        adata,
    ) in data.try_convert_to_anndata_from_disk(
        intensities_files,
        *data_file_lists,
        neighbors_files=neighbors_files,
        # panel=panel,
        # image_info=image_info,
    ):
        if concatenate:
            adatas[img_file_name] = adata
        else:
            anndata_file = io._as_path_with_suffix(
                Path(anndata_file_or_dir) / img_file_name, ".adata"
            )
            anndata_file = write_anndata(adata, anndata_file)
            logger.info(anndata_file)
            del adata
    if concatenate:
        adata = anndata_concat(
            adatas, merge="first", label="Image", index_unique=" in "
        )
        obs_cols = list(adata.obs.columns)
        obs_cols.insert(0, obs_cols.pop(obs_cols.index("Image")))
        adata.obs = adata.obs.loc[:, obs_cols]
        write_anndata(adata, anndata_file_or_dir, ignore_suffix=True)
        logger.info(anndata_file_or_dir)



import click

from ..._cli.utils import OrderedClickGroup
from .external import external_cmd_group
from .imc import imc_cli_available, imc_cmd_group


@click.group(
    name="preprocess",
    cls=OrderedClickGroup,
    help="Extract and preprocess images from raw data",
)
def preprocess_cmd_group():
    pass


if imc_cli_available:
    preprocess_cmd_group.add_command(imc_cmd_group)

preprocess_cmd_group.add_command(external_cmd_group)



import click

from ..._cli.utils import OrderedClickGroup
from .cellprofiler import cellprofiler_cmd_group
from .deepcell import deepcell_cli_available, deepcell_cmd


@click.group(
    name="segment",
    cls=OrderedClickGroup,
    help="Perform image segmentation to create object masks",
)
def segment_cmd_group():
    pass


segment_cmd_group.add_command(cellprofiler_cmd_group)
if deepcell_cli_available:
    segment_cmd_group.add_command(deepcell_cmd)
